# pat.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""


import copy
import random

import mma.MMA.notelen
import mma.MMA.translate
import mma.MMA.midi
import mma.MMA.midiC
import mma.MMA.mdefine
import mma.MMA.volume
import mma.MMA.alloc
import mma.MMA.seqrnd
import mma.MMA.truncate
import mma.MMA.ornament
import mma.MMA.trigger
import mma.MMA.rpitch
import mma.MMA.debug

from . import gbl
from mma.MMA.common import *
from mma.MMA.miditables import NONETONE

pats = {}        # Storage for all pattern defines

class Pgroup:
    pass


defaultDrum = 0
defaultVoice = 0
currentChord = None

def getIncDec(v):
    """ See if 'v' starts with -/+. Strip and return. """

    if v.startswith('-'):
        incr = -1
        v = v[1:]
    elif v.startswith('+'):
        incr = 1
        v = v[1:]
    else:
        incr = 0

    return incr, v


def getIncDecValue(orig, inc, new):
    """ Set value based on original, increment and new values. """

    if inc == -1:
        return orig - new
    elif inc == 1:
        return orig + new
    else:
        return new


def getRndPair(l, trk, min, max):
    """ Parse off a pair (or single) value for random settings (rvolume, etc.). """

    if not l:
        error("%s: expecting a value or value pair." % trk)

    if l.count(',') > 1:
       error("%s: Random pairs can only have one ','." % trk)

    if l.count(',') == 1:
        n1, n2 = l.split(',')
        try:
            n1 = int(n1)
            n2 = int(n2)
        except:
            error("%s: Expecting integers, not '%s' or '%s'." % (trk, n1, n2))
    else:
        try:
            n1 = int(l)
        except:
            error("%s: Expecting integer, not '%s'." % (trk, n1))
        n2 = n1 * -1

    if n2 < n1:
        n1, n2 = n2, n1

    if n1 < min or n1 > max or n2 < min or n2 > max:
        error("%s: Max range is %s to %s." % (trk, min, max))

    return (n1, n2)


class PC:
    """ Pattern class.

        Define classes for processing drum, chord, arp, and chord track.
        These are mostly the same, so we create a base class and derive
        the others from it.

        We have a class for each track type. They are all derived
        from the base PC class. Some classes have their own __init__().
        They call back here after doing their own special stuff.

    """

    def __init__(self, nm):

        self.inited = 0
        self.name = nm
        self.channel = 0
        self.grooves = {}
        self.saveVols = {}
        self.ssvoice = -1  # Track the voice set for the track
        self.smidiVoice = ()  # Track MIDIVoice cmds to avoid dups
        self.midiSent = 0   # if set, MIDICLEAR invoked.

        """ Midi commands like Pan, Glis, etc. are stacked until musical
            data is actually written to the track. Each item in
            the midiPending list is a name (PAN, GLIS, etc), timeoffset, value.
        """

        self.midiPending = []

        self.riff = []

        self.disable = 0
        self.sticky = False
        self.clearSequence()

        self.nextVolume = None
        self.cVolume = 127  # assume device is set to max.
        self.inited = 1

        # Double test is needed! We disable a track only
        # when there is a list of tracks which are NOT
        # disabled. The variable name is slightly misleading.
        if gbl.muteTracks and nm not in gbl.muteTracks:
            self.disable = 1


    ##########################################
    ## These are called from process() to set options

    def setCompress(self, ln):
        """ set/unset the compress flag. """

        ln = lnExpand(ln, '%s Compress' % self.name)

        tmp = []

        for n in ln:

            if n.upper() in ("ON", "TRUE"):
                n = "1"
            if n.upper() in ("OFF", "FALSE"):
                n = "0"
            if n not in ("0", "1"):
                error("%s Compress: Argument 0, 1, True, False, not '%s'." % (self.name, n))

            if n == '1':
                n = 1
            else:
                n = 0
            if n and self.vtype == 'CHORD' and self.voicing.mode:
                vwarn = 1

            tmp.append(n)

        self.compress = seqBump(tmp)

        if self.vtype not in ("CHORD", "ARPEGGIO"):
            warning("Compress is ignored in %s tracks" % self.vtype)

        if mma.MMA.debug.debug:
            mma.MMA.debug.trackSet(self.name, "Compress")

    def setRange(self, ln):
        """ set range. """

        ln = lnExpand(ln, '%s Range' % self.name)

        tmp = []

        for n in ln:
            n = stof(n)
            if n == 0:
                n = 1
            if n <= 0 or n >= 6:
                error("Range %s out-of-range; must be between 0..6, not %s" % (self.name, n))

            tmp.append(n)

        self.chordRange = seqBump(tmp)

        if self.vtype not in ("SCALE", "ARPEGGIO", "ARIA"):
            warning("%s Range: ignored in '%s' tracks" % (self.name, self.vtype))

        if self.vtype == 'ARIA':
            self.restart()

        if mma.MMA.debug.debug:
            mma.MMA.debug.trackSet(self.name, "Range")

    def setVoicing(self, ln):
        """ set the Voicing Mode options.

            This is an error trap stub. The real code is in patChord.py
            and patSolo.py (settings are only valid for those types).
        """

        error("Voicing is not supported for %s tracks" % self.vtype)

    def setForceOut(self):
        """ Set the force output flag. This does 2 things: assigns
            a midi channel and sends the voicing setting to the track.
        """

        if not self.channel:
            self.setChannel()
        self.clearPending()

        self.insertVoice(gbl.seqCount)

    def setDelay(self, ln):
        """ Set a delay for this track. """

        ln = lnExpand(ln, '%s Delay' % self.name)

        tmp = []
        for n in ln:
            if n[0] == '-':
                mul = -1
                n = n[1:]
            elif n[0] == '+':
                mul = 1
                n = n[1:]
            else:
                mul = 1
            tmp.append(mma.MMA.notelen.getNoteLen(n) * mul)

        self.delay = seqBump(tmp)

        if mma.MMA.debug.debug:
            mma.MMA.debug.trackSet(self.name, "Delay")

    def setDupRoot(self, ln):
        """ set/unset root duplication.

            This is a stub. Only valid for CHORDs and that is where the code is."""

        warning("RootDup has no effect in %s tracks" % self.vtype)

    def setChords(self, ln):
        """ Set track specific chord(s). """

        # An empty arg. or a single '-' turns off this stuff.
        if not ln:
            self.chord = seqBump([[]])
            return

        # Bracket the chord data. It can be passed as "C D E" or
        # "{ C D E } {i ii iii}". Also permitted is an empty {} or
        # a single "/" (repeat last) with or without {}.

        # make the entire line into a long string.
        ln = ' '.join(ln)

        clist = []
        while 1:   # loop on entire string
            if not ln:
                break

            if ln[0] == '/':  # handle single repeat
                if clist:
                    clist.append(clist[-1])
                    ln = ln[1:].strip()
                    continue
                else:
                    error("%s CHORDS: cannot start with a '/'." % self.name)

            if ln[0] != '{':
                ln = '{ ' + ln + '}'
                continue

            ln, s = pextract(ln, '{', '}', onlyone=True)
            if not s:
                error("%s CHORDS: Multiple bar chords need {}s." % self.name)

            s = s[0].split()
            if len(s) == 1 and s[0] == '/':    # this would be a {/}
                ln = '/' + ln
                continue    # loop back and handle repeat above

            if s and s[0] == '/':
                error("%s CHORDS: Chord list cannot start with '/'." % self.name)

            clist.append(s)

        if len(clist) > gbl.seqSize:
            warning("%s CHORDS: Too many bars(%s) for seq size(%s). Chord list truncated."
                    % (self.name, len(clist), gbl.seqSize))

        self.chord = seqBump(clist)

        if mma.MMA.debug.debug:
            mma.MMA.debug.trackSet(self.name, "Chords")

    def setChordLimit(self, ln):
        """ set/unset the chordLimit flag. """

        n = stoi(ln, "%s Limit argument for must be a value" % self.name)

        if n < 0 or n > 8:
            error("%s Limit %s out-of-range; must be 0 to 8" % (self.name, n))

        self.chordLimit = n

        if self.vtype not in ("CHORD", "ARPEGGIO"):
            warning("%s Limit is ignored in %s tracks" % (self.name, self.vtype))

        if mma.MMA.debug.debug:
            mma.MMA.debug.trackSet(self.name, "Limit")

    def setChannel(self, ln=None):
        """ Set the midi-channel number for a track.

        - Checks for channel duplication
        - Auto assigns channel number if ln==''


        If no track number was passed, then we try to
        auto-alloc a track. First, we see if a preference
        was set via MidiChPref. If these is no preference,
        or if the preferred channel is already allocated
        we go though the list, top to bottom, to find
        an available channel.
        """

        if not ln:
            try:
                c = gbl.midiChPrefs[self.name]
            except:
                c = 0

            if not c or gbl.midiAvail[c]:
                c = -1
                for a in range(16, 0, -1):
                    if a != 10 and not gbl.midiAvail[a]:
                        c = a
                        break

            if c < 0:
                error("No MIDI channel is available for %s, "
                      "Try CHShare or Delete unused tracks" % self.name)

        else:
            c = stoi(ln, "%s Channel assignment expecting Value, not %s" %
                    (self.name, ln))

            if c < 0 or c > 16:
                error("%s Channel must be 0..16, not %s" % (self.name, ln))

        if c == 10:
            if self.vtype == 'DRUM':
                pass
            elif self.vtype in ('SOLO', 'MELODY') and self.drumType:
                pass
            else:
                error("Channel 10 is reserved for DRUM, not %s" % self.name)

        if self.vtype == 'DRUM' and c != 10:
            error("DRUM tracks must be assigned to channel 10")

        # Disable the channel.

        if c == 0:
            if gbl.midiAvail[self.channel]:
                gbl.midiAvail[self.channel] -= 1
            s = "%s channel disabled" % self.name
            if gbl.midiAvail[self.channel]:
                s += " Other tracks are still using channel %s" % self.channel
            else:
                s += " Channel %s available" % self.channel
            warning(s)
            self.channel = 0
            self.disable = 1
            return

        if c != 10:
            for a, tr in gbl.tnames.items():
                if a == self.name:    # okay to reassign same number
                    continue

                if tr.channel == c:
                    error("Channel %s is assigned to %s" % (c, tr.name))

        self.channel = c
        if not self.name in gbl.midiAssigns[c]:
            gbl.midiAssigns[c].append(self.name)

        gbl.midiAvail[c] += 1

        if not c in gbl.mtrks:
            gbl.mtrks[c] = mma.MMA.midi.Mtrk(c)
            if mma.MMA.debug.debug:
                dPrint("MIDI channel %s buffer created" % c)

        # If a track name has not be set via MidiTName (this will be
        # queued since the track just got a midi assignment here) we
        # put a default track name into the queue.

        if c != 10:
            f = 0
            for a, i in enumerate(self.midiPending):
                if i[0] == 'TNAME':
                    f = 1
            if not f:
                self.midiPending.append(('TNAME', 0, self.name.title()))

        if mma.MMA.debug.debug:
            dPrint("MIDI Channel %s assigned to %s" % (self.channel, self.name))

    def setChShare(self, ln):
        """ Share midi-channel setting. """

        if self.channel:    # If channel already assigned, ignore
            warning("Channel for %s has previously been assigned "
                    "(can't ChShare)" % self.name)
            return

        """ Get name of track to share with and make sure it exists.
            If not, trackAlloc() will create the track. Do some
            sanity checks and ensure that the shared track has
            a channel assigned.
        """

        sc = ln.upper()

        mma.MMA.alloc.trackAlloc(sc, 1)

        if not sc in gbl.tnames:
            error("Channel '%s' does not exist. No such name" % sc)

        if sc == self.name:
            error("%s can't share MIDI channel with itself" % sc)

        if not gbl.tnames[sc].channel:
            gbl.tnames[sc].setChannel()

        schannel = gbl.tnames[sc].channel

        if not schannel:
            error("CHShare attempted to assign MIDI channel for %s, but "
                  "none avaiable" % self.name)

        """ Actually do the assignment. Also copy voice from
            base track to this one ... it's going to use that voice anyway?
        """

        self.channel = schannel
        self.voice = gbl.tnames[sc].voice[:]

        # Update the avail. lists

        gbl.midiAssigns[self.channel].append(self.name)
        gbl.midiAvail[self.channel] += 1

    def setSticky(self, ln):
        """ Set track as sticky. Sticky tracks are ignored by groove commands. """

        if len(ln) != 1:
            error("%s Sticky needs single argument ('True/False')." % self.name)
         
        arg = ln[0].upper()
        if arg in ("TRUE", "ON", "1"):
            self.sticky = True

        elif arg in ("FALSE", "OFF", "0"):
            self.sticky = False 

        else:
            error("%s Sticky: '%s' is not a valid option." % (self.name, arg))

        if mma.MMA.debug.debug:
            mma.MMA.debug.trackSet(self.name, "Sticky")

        
    def setStrum(self, ln):
        """ Set Strum time. """

        ln = lnExpand(ln, '%s Strum' % self.name)
        tmp = []

        for n in ln:
            if ',' in n:
                a, b = n.split(',', 1)
                a = stoi(a, "Argument for %s Strum must be an integer" % self.name)
                b = stoi(b, "Argument for %s Strum must be an integer" % self.name)
            else:
                a = stoi(n, "Argument for %s Strum must be an integer" % self.name)
                b = a

            if a > b:
                b, a = a, b

            if a < -300 or a > 300 or b < -300 or b > 300:
                error("STRUM: %s out-of-range. All values must be -300..300" % n)

            if a == 0 and b == 0:
                tmp.append(None)
            else:
                tmp.append((a, b))

        self.strum = seqBump(tmp)

        if self.vtype == "DRUM":
            warning("Strum has no effect in %s tracks" % self.name)

        if mma.MMA.debug.debug:
            mma.MMA.debug.trackSet(self.name, "Strum")

    def setStrumAdd(self, ln):
        """ Additional options for strum. """

        ln = lnExpand(ln, '%s StrumAdd' % self.name)
        tmp = []

        for i in ln:
            v = stoi(i, "Argument for %s StrumAdd must be an integer" % self.name)
            tmp.append(v)

        self.strumAdd = seqBump(tmp)

        if mma.MMA.debug.debug:
            mma.MMA.debug.trackSet(self.name, "StrumAdd")

    def getStrum(self, sc):
        """ Returns the strum factor. Note that if strum==(0,0) a 0 is returned."""
        

        if not self.strum[sc]:
            return 0
        
        a, b = self.strum[sc]  # split out tuple. Always 2 values (or empty)
        if a == b:
            return a
        
        return random.randint(a, b)

    def setTone(self, ln):
        """ Set Tone. Error trap, only drum tracks have tone. """
        
        error("Tone command not supported for %s track" % self.name)

    def setOn(self):
        """ Turn ON track. """

        if gbl.muteTracks and self.name not in gbl.muteTracks:
            warning("Attempt to enable muted track %s ignored." % self.name)

        else:
            self.disable = 0
            self.ssvoice = -1

        if mma.MMA.debug.debug:
            dPrint("%s Enabled" % self.name)

    def setOff(self):
        """ Turn OFF track. """

        self.disable = 1

        if mma.MMA.debug.debug:
            dPrint("%s Disabled" % self.name)

    def setRVolume(self, ln):
        """ Set the volume randomizer for a track. """

        msg = "%s Rvolume" % self.name
        ln = lnExpand(ln, msg)
        tmp = []
        for n in ln:
            n1, n2 = getRndPair(n, msg, -100, 100)
            tmp.append([n1 / 100., n2 / 100.])

        self.rVolume = seqBump(tmp)

        if mma.MMA.debug.debug:
            mma.MMA.debug.trackSet(self.name, "RVolume")

         
    def setRSkip(self, ln):
        """ Set the note random skip factor for a track. """

        msg = "%s RSkip" % self.name
        ln, options = opt2pair(ln)

        if not ln:
            error("%s: Needs weighting list and optional beats=[]." % msg)

        self.rSkipBeats = []
        for cmd, opt in options:
            if cmd.upper() == 'BEATS':
                beats = opt.split(',')
                if not beats:
                    error("%s Beats: Expecting list of affected beats." % msg)
                self.rSkipBeats = []
                for b in beats:
                    b = stof(b)
                    if b < 1:
                        error("%s: Beats must be => 1, not '%s'" % (msg, b))
                    if b >= gbl.QperBar + 1:
                        error("%s: Beats must less than %s, not '%s'" % (msg, gbl.QperBar + 1, b))
                    self.rSkipBeats.append(int((b - 1) * gbl.BperQ))

            else:
                error("%s: Unknown option %s." % (msg, cmd))

        ln = lnExpand(ln, msg)
        tmp = []

        for n in ln:
            n = stoi(n, "%s: Expecting integer" % msg)

            if n < 0 or n > 99:
                error("%s: arg must be 0..99" % msg)

            tmp.append(n / 100.)

        self.rSkip = seqBump(tmp)

        if mma.MMA.debug.debug:
            mma.MMA.debug.trackSet(self.name, "RSkip")


    def setRDuration(self, ln):
        """ Set the duration randomizer for a track. """

        msg = "%s RDuration" % self.name
        ln = lnExpand(ln, msg)
        tmp = []

        for n in ln:
            n1, n2 = getRndPair(n, msg, -100, 100)
            tmp.append([n1 / 100., n2 / 100.])

        self.rDuration = seqBump(tmp)

        if mma.MMA.debug.debug:
            mma.MMA.debug.trackSet(self.name, "RDuration")


    def setRTime(self, ln):
        """ Set the timing randomizer for a track. """

        msg = "%s RTime" % self.name
        ln = lnExpand(ln, msg)
        tmp = []

        for n in ln:
            n1, n2 = getRndPair(n, msg, -100, 100)
            tmp.append([n1, n2])

        self.rTime = seqBump(tmp)

        if mma.MMA.debug.debug:
            mma.MMA.debug.trackSet(self.name, "RTime")

    def setDirection(self, ln):
        """ Set scale direction. """

        ln = lnExpand(ln, "%s Direction" % self.name)
        tmp = []

        for n in ln:
            n = n.upper()
            if not n in ('UP', 'DOWN', 'BOTH', 'RANDOM'):
                error("Unknown %s Direction '%s'" % (self.name, n))
            tmp.append(n)

        self.direction = seqBump(tmp)

        if self.vtype == 'SCALE':
            self.lastChord = None
            self.lastNote = -1

        if mma.MMA.debug.debug:
            mma.MMA.debug.trackSet(self.name, "Direction")

    def setScaletype(self, ln):
        """ Set scale type.

            This is a error stub. The real code is in the permitted track code.
        """

        warning("ScaleType has no effect in %s tracks" % self.vtype)

    def setInvert(self, ln):
        """ Set inversion for track.

            This can be applied to any track,
            but has no effect in drum tracks. It inverts the chord
            by one rotation for each value.
        """

        msg = "%s Invert" % self.name
        ln = lnExpand(ln, msg)

        tmp = []

        for n in ln:
            if not ',' in n:
                try:
                    n1 = n2 = int(n)
                except:
                    error("Invert %s doesn't recognize '%s'." % (self.name, n))
            else:
                n1, n2 = getRndPair(n, msg, -10, 10)
            if n1 == 0 and n2 == 0:
                tmp.append([])
            else:
                tmp.append([n1, n2])

        if self.vtype == 'CHORD' and self.voicing.mode and any(tmp):
            warning("%s: Setting both Voicing Mode and Invert is not a good idea" % 
                    self.name)

        if self.vtype not in ("CHORD", "ARPEGGIO"):
            warning("Invert is ignored in %s tracks" % self.vtype)

        self.invert = seqBump(tmp)

        if mma.MMA.debug.debug:
            mma.MMA.debug.trackSet(self.name, "Invert")

    def setOctave(self, ln):
        """ Set the octave for a track. Use 0-10, -x, +x. """

        ln = lnExpand(ln, '%s Octave' % self.name)
        ln = seqBump(ln)  # Needed for +/- values to complete
        tmp = []
        i = 0
        for n in ln:
            incr, n = getIncDec(n)
            n = stoi(n, "Argument for %s Octave must be an integer" % self.name)
            n = getIncDecValue(self.octave[i] // 12, incr, n)

            if n < 0 or n > 10:
                error("Octave %s out-of-range; must be 0..10" % n)
            tmp.append(n * 12)
            i += 1

        self.octave = seqBump(tmp)

        if mma.MMA.debug.debug:
            dPrint("Set %s Octave to: %s" % 
                  (self.name, ' '.join([str(i // 12) for i in self.octave])))


    def setMOctave(self, ln):
        """ Set the octave for a track based on MIDI octave values. Use -1..9. """

        ln = lnExpand(ln, '%s MOctave' % self.name)
        tmp = []

        for n in ln:
            n = stoi(n, "Argument for %s MOctave must be an integer" % self.name)
            if n < -1 or n > 9:
                error("MOctave %s out-of-range; must be -1..9" % n)
            tmp.append((n + 1) * 12)

        self.octave = seqBump(tmp)

        if mma.MMA.debug.debug:
            dPrint("Set %s MOctave to: %s" % 
                  (self.name, ' '.join([str((i // 12) - 1) for i in self.octave])))


    def setSpan(self, start, end):
        """ Set span.

            Note: The start/end parm has been verified in parser.

        """

        if self.vtype == 'DRUM':
            warning("Span has no effect in Drum tracks")

        self.spanStart = start
        self.spanEnd = end

        if mma.MMA.debug.debug:
            dPrint("Set %s Span to %s...%s" % (self.name, self.spanStart, self.spanEnd))

    def setSeqSize(self):
        """ Expand existing pattern list.

            Track functions may have their own and do a callback.
        """

        self.sequence = seqBump(self.sequence)
        if self.midiVoice:
            self.midiVoice = seqBump(self.midiVoice)
        if self.midiSeq:
            self.midiSeq = seqBump(self.midiSeq)
        self.invert = seqBump(self.invert)
        self.artic = seqBump(self.artic)
        self.chord = seqBump(self.chord)
        self.volume = seqBump(self.volume)
        self.voice = seqBump(self.voice)
        self.rVolume = seqBump(self.rVolume)
        self.rSkip = seqBump(self.rSkip)
        self.rTime = seqBump(self.rTime)
        self.rDuration = seqBump(self.rDuration)
        self.seqRndWeight = seqBump(self.seqRndWeight)
        self.strum = seqBump(self.strum)
        self.strumAdd = seqBump(self.strumAdd)
        self.octave = seqBump(self.octave)
        self.harmony = seqBump(self.harmony)
        self.harmonyOnly = seqBump(self.harmonyOnly)
        self.harmonyVolume = seqBump(self.harmonyVolume)
        self.direction = seqBump(self.direction)
        self.delay = seqBump(self.delay)
        self.scaleType = seqBump(self.scaleType)
        self.compress = seqBump(self.compress)
        self.chordRange = seqBump(self.chordRange)
        self.dupRoot = seqBump(self.dupRoot)
        self.unify = seqBump(self.unify)
        self.accent = seqBump(self.accent)

    def setVoice(self, ln):
        """ Set the voice for a track.

            Note, this just sets flags, the voice is set in bar().
            ln[] is not nesc. set to the correct length.

            the voice can be gm-string, user-def-string, or value.
            A value can be xx.yy.vv, yy.vv or vv
        """

        ln = lnExpand(ln, '%s Voice' % self.name)
        tmp = []

        for n in ln:
            voc = mma.MMA.midiC.decodeVoice(n)
            tmp.append(voc)

        self.voice = seqBump(tmp)

        if self.channel and len(gbl.midiAssigns[self.channel]) > 1:
            a = []
            for n in gbl.midiAssigns[self.channel]:
                if n != self.name:
                    a.append(n)

            if len(a) > 1:      # a bit for grammar
                t1 = "Tracks"
                v1 = "are"
            else:
                t1 = "Track"
                v1 = "is"

            warning("%s %s %s shared with %s. Changing voice may create conflict"
                    % (t1, ', '.join(a), v1, self.name))

        if mma.MMA.debug.debug:
            dPrint("Set %s Voice to: %s" % 
                  (self.name, ' '.join([mma.MMA.midiC.valueToInst(i) for i in self.voice])))


    def setMidiClear(self, ln):
        """ Set MIDIclear sequences. """

        if ln[0] in 'zZ-':
            self.midiClear = None
        else:
            self.midiClear = mma.MMA.mdefine.mdef.get(ln[0])

        if mma.MMA.debug.debug:
            dPrint("%s MIDIClear: %s" % (self.name, self.midiSeqFmt(self.midiClear)))

    def doMidiClear(self):
        """ Reset MIDI settings. """

        if self.midiSent:
            if not self.midiClear:
                warning("%s: Midi data has been inserted with MIDIVoice/Seq "
                        "but no MIDIClear data is present" % self.name)

            else:
                for i in self.midiClear:
                    gbl.mtrks[self.channel].addCtl(gbl.tickOffset, i[1])

            self.midiSent = 0

    def doChannelReset(self):
        """ Reset the midi voicing to 'sane'. Called when track ended.

            Note that standard MIDI voicing is left alone, only extended
            voicing is affected.

        """

        if self.ssvoice > 127 and self.ssvoice != NONETONE:
            gbl.mtrks[self.channel].addProgChange(gbl.tickOffset, 0, self.ssvoice)

    def setMidiSeq(self, ln):
        """ Set a midi sequence for a track.

        This is sent for every bar. Syntax is:
        <beat> <ctrl> hh .. ; ...

        or a single '-' to disable.
        """

        """ lnExpand() works here! The midi data has been converted to
             pseudo-macros already in the parser. """

        ln = lnExpand(ln, "%s MidiSeq" % self.name)

        seq = []
        for a in ln:
            if a in 'zZ-':
                seq.append(None)
            else:
                seq.append(mma.MMA.mdefine.mdef.get(a.upper()))

        if seq.count(None) == len(seq):
            self.midiSeq = []
        else:
            self.midiSeq = seqBump(seq)

        if mma.MMA.debug.debug:
            dPrint("%s MIDISeq: %s" % 
                  (self.name, ' '.join(['{ %s }' % self.midiSeqFmt(l) for l in seq])))


    def setMidiVoice(self, ln):
        """ Set a MIDI sequence for a track.

        This is sent whenever we send a VOICE. Syntax is:
        <beat> <ctrl> hh .. ; ...

        or a single '-' to disable.
        """

        # lnExpand() works here! The midi data has been converted to
        # pseudo-macros already in the parser. 
        ln = lnExpand(ln, '%s MIDIVoice' % self.name)

        seq = []
        for a in ln:
            if a in 'zZ':  # some bars in seq might not want midi stuff
                seq.append(None)
            else:
                seq.append(mma.MMA.mdefine.mdef.get(a.upper()))

        if seq.count(None) == len(seq):  # see if all bars in seq are None
            self.midiVoice = []          # yes, kill the whole voice thing
        else:
            self.midiVoice = seqBump(seq)  # set voice for all bars in seq
 
        if mma.MMA.debug.debug:
            dPrint("%s MIDIVoice: %s" %
                  (self.name, ' '.join(['{ %s }' % self.midiSeqFmt(l) for l in seq])))


    def midiSeqFmt(self, lst):
        """ Used by setMidiVoice/Clear/Seq for debugging format. """

        if lst is None:
            return ''
        ret = ''
        for i in lst:
            ret += "%s %s 0x%02x ; " % (i[0],
                          mma.MMA.midiC.valueToCtrl(i[1][0]), i[1][1])
        return ret.rstrip("; ")

    def setVolume(self, ln):
        """ Set the volume for a pattern.
            ln - list of volume names (pp, mf, etc)
            ln[] not nesc. correct length
        """

        ln = lnExpand(ln, '%s Volume' % self.name)
        ln = seqBump(ln)  # needed for -/+ setting to apply to all seqs

        tmp = [None] * len(ln)
        self.futureVols = []   # clear off dangling (de)cresc for voice.


        for i, n in enumerate(ln):
            a = mma.MMA.volume.calcVolume(n, self.volume[i])
            if self.vtype == 'DRUM' or \
                    (self.vtype in ('MELODY', 'SOLO') and self.drumType):
                a = mma.MMA.translate.drumVolTable.get(self.toneList[i], a)
            else:
                a = mma.MMA.translate.voiceVolTable.get(self.voice[i], a)
            tmp[i] = a

        self.volume = seqBump(tmp)

        if mma.MMA.debug.debug:
            dPrint("Set %s Volume to: %s" %
                  (self.name, ' '.join([str(int(a * 100))  for a in self.volume])))

    def setCresc(self, dir, ln):
        """ Set Crescendo for a track.     """

        if len(ln) == 3:
            self.setVolume([ln[0]])
            ln = ln[1:]

        if len(ln) != 2:
            error("Cresc expecting 2 or 3 args.")

        vol = self.volume[0]

        if self.volume.count(vol) != len(self.volume):
            warning("(De)Crescendo being used with track with variable sequence volumes")

        self.futureVols = mma.MMA.volume.fvolume(dir, vol, ln)

        if mma.MMA.debug.debug:
            dPrint("Set %s Cresc to: %s" % 
                  (self.name, ' '.join([str(int(a * 100)) for a in self.futureVols])))


    def setSwell(self, ln):
        """ Set a swell (cresc<>decresc) for track. """

        if len(ln) == 3:            # 3 args, 1st is intial setting
            self.setVolume([ln[0]])
            ln = ln[1:]

        if len(ln) != 2:
            error("%s Swell expecting 2 or 3 args." % self.name)

        if self.volume.count(self.volume[0]) != len(self.volume):
            warning("%s Swell being used with track with variable sequence volumes."
                    % self.name)

        count = stoi(ln[1])
        if count < 2:
            error("%s Swell bar count must be 2 or greater." % self.name)

        count += 1
        c = count // 2
        if count % 2:   # even number of bars (we bumped count)
            offset = 1
            c += 1
        else:           # odd number
            offset = 0
        c = str(c)

        self.futureVols = mma.MMA.volume.fvolume(0, self.volume[0],
                                             [ln[0], c])

        self.futureVols.extend(mma.MMA.volume.fvolume(0, self.futureVols[-1],
             [str(int(self.volume[0] * 100)), c])[offset:])

        if mma.MMA.debug.debug:
            dPrint("Set %s Swell to: %s" % 
                  (self.name, ' '.join([str( int(a * 100)) for a in self.futureVols])))

    def setMallet(self, ln):
        """ Mallet (repeat) settngs. """

        if self.vtype == "PLECTRUM":
            warning("%s: Mallet has no effect, ignored." % self.name)
            return

        for l in ln:
            try:
                mode, val = l.upper().split('=')
            except:
                error("%s Mallet: each option must contain a '=', not '%s'" % (self.name, l))

            if mode == 'RATE':
                if val == '0' or val.upper() == 'NONE':
                    self.mallet = 0
                else:
                    self.mallet = mma.MMA.notelen.getNoteLen(val)

            elif mode == 'DECAY':
                val = stof(val, "Mallet Decay must be a value, not '%s'" % val)

                if val < -50 or val > 50:
                    error("%s Mallet: Decay rate must be -50..+50" % self.name)

                self.malletDecay = val / 100.

            else:
                error("%s Mallet: %s is not a option." % (self.name, mode))

        if mma.MMA.debug.debug:
            dPrint("%s Mallet Rate:%s Decay:%s" %
                (self.name, self.mallet, self.malletDecay))

    def setAccent(self, ln):
        """ Set the accent values. This is a list of lists, a list for each seq. """

        tmp = []

        """ 2 ways to call this:

            Track Accent 1 20 3 -10   -- Sets accent for all bars in sequence to
                   the same values. The parser normalizes this by adding a {}
                   around the args.
            Track Accent {1 20 3 -10} -- Same as above
            Track Accent {1 20} / {/} {3 20} -- Different for each bar in seq.

            At this point ln is a string. We extract each set of {}s and build a
            new ln[].

            Note that the "/" can or not have {}s.

            An empty string resets to no accents.
        """

        ln = ' '.join(ln)
        if not ln.startswith('{'):
            ln = '{' + ln + "}"

        l = []
        while ln:
            if ln[0] == "/":  # / not delimited with {}
                ln = "{/}" + ln[1:]

            if ln[0][0] != '{':
                error("%s Accent: Unmatched or missing '{} delimters" % self.name)

            ln, b = pextract(ln, "{", "}", onlyone=True)

            if len(b) == 1 and b[0] == '/':   # convert ['/'] to '/' for lnExpand()
                l.append('/')
            else:
                l.append(b[0].split())

        ln = lnExpand(l, '%s Accent' % self.name)

        for l in ln:
            tt = []
            if len(l) // 2 * 2 != len(l):
                error("Use: %s Accent Beat Percentage [...]" % self.name)

            for b, v in zip(l[::2], l[1::2]):
                b = self.setBarOffset(b)
                v = stoi(v, "Beat offset must be a value, not '%s'" % v)
                if v < -100 or v > 100:
                    error("Velocity adjustment (as percentage) must "
                          "be -100..100, not '%s'" % v)

                tt.append((b, v / 100.))
            tmp.append(tt)

        self.accent = seqBump(tmp)

        if mma.MMA.debug.debug:
            accList = []
            for s in self.accent:
                accList.append('{')
                for b, v, in s:
                    accList.append("%s %s" % (1 + (b / float(gbl.BperQ)), int(v * 100)))
                accList.append('}')
            dPrint("%s Accent: %s" % (self.name, ' '.join(accList)))

    def setArtic(self, ln):
        """ Set the note articuation value. """

        ln = lnExpand(ln, '%s Articulate' % self.name)
        ln = seqBump(ln)
        tmp = []
        incr = 0
        i = 0

        for n in ln:
            incr, n = getIncDec(n)
            a = stoi(n, "Expecting value in articulation setting")
            a = getIncDecValue(self.artic[i], incr, a)

            if self.vtype is 'PLECTRUM':
                if a < 0 or a > 500:
                    error("%s: Articulation setting must be 0 .. 500 (midi ticks), not %s" %
                          (self.name, a))
            else:
                if a < 1 or a > 500:
                    error("%s: Articulation setting must be 1..500, not %s" % (self.name, a))

                if a > 200:
                    warning("%s: Articulation '%s' is a large value." % (self.name, a))

            tmp.append(a)
            i += 1

        self.artic = seqBump(tmp)

        if mma.MMA.debug.debug:
            dPrint("Set %s Articulate to: %s" % 
                  (self.name, ' '.join([str(a) for a in self.artic])))


    def setUnify(self, ln):
        """ Set unify. """

        ln = lnExpand(ln, "%s Unify" % self.name)
        tmp = []

        for n in ln:
            n = n.upper()
            if n in ('ON', 'TRUE', '1'):
                tmp.append(1)
            elif n in('OFF', 'FALSE', '0'):
                tmp.append(0)
            else:
                error("Unify accepts On/True/1 or Off/False/0, not %s." % n)

        self.unify = seqBump(tmp)

        if mma.MMA.debug.debug:
            dPrint("Set %s Unify to: %s" %
              (self.name, ' '.join([str(a) for a in self.unify])))

    ##################################################
    ## Save/restore grooves

    def saveGroove(self, gname):
        """ Define a groove.

            Called by the 'DefGroove Name'. This is called for
            each track.

            If 'gname' is already defined it is overwritten.

            Note that tracks may have their own function for local/specific variables.
            They call here first to create storage, then do their own.
        """

        self.grooves[gname] = {
            'ACCENT': self.accent[:],
            'ARTIC': self.artic[:],
            'CHORD': copy.deepcopy(self.chord),
            'COMPRESS': self.compress[:],
            'DELAY': self.delay[:],
            'DIR': self.direction[:],
            'DUPROOT': copy.deepcopy(self.dupRoot),
            'HARMONY': (self.harmony[:], self.harmonyOnly[:], self.harmonyVolume[:]),
            'INVERT': copy.deepcopy(self.invert),
            'LIMIT': self.chordLimit,
            'RANGE': self.chordRange[:],
            'OCTAVE': self.octave[:],
            'RSKIP': self.rSkip[:],
            'RSKIPBEATS': self.rSkipBeats[:],
            'RTIME': copy.deepcopy(self.rTime),
            'RDURATION': copy.deepcopy(self.rDuration),
            'RVOLUME': copy.deepcopy(self.rVolume),
            'RPITCH': copy.deepcopy(self.rPitch),
            'SCALE': self.scaleType[:],
            'SEQ': self.sequence[:],
            'SEQRND': self.seqRnd,
            'SEQRNDWT': self.seqRndWeight[:],
            'STRUM': self.strum[:],
            'STRUMADD': self.strumAdd[:],
            'VOICE': self.voice[:],
            'VOLUME': self.volume[:],
            'UNIFY': self.unify[:],
            'MIDISEQ': self.midiSeq[:],
            'MIDIVOICE': self.midiVoice[:],
            'MIDICLEAR': self.midiClear[:],
            'SPAN': (self.spanStart, self.spanEnd),
            'MALLET': (self.mallet, self.malletDecay),
            'ORNAMENT': copy.deepcopy(self.ornaments)}

    def restoreGroove(self, gname):
        """ Restore a defined groove.

            Some tracks will override to restore their own variables. They
            then call back to this to finish the job.
        """

        self.doMidiClear()

        g = self.grooves[gname]

        self.sequence = g['SEQ']
        self.volume = g['VOLUME']
        self.accent = g['ACCENT']
        self.rTime = g['RTIME']
        self.rDuration = g['RDURATION']
        self.rVolume = g['RVOLUME']
        self.rPitch = g['RPITCH']
        self.rSkip = g['RSKIP']
        self.rSkipBeats = g['RSKIPBEATS']
        self.strum = g['STRUM']
        self.strumAdd = g['STRUMADD']
        self.octave = g['OCTAVE']
        self.voice = g['VOICE']
        self.harmony, self.harmonyOnly, self.harmonyVolume = g['HARMONY']
        self.direction = g['DIR']
        self.delay = g['DELAY']
        self.scaleType = g['SCALE']
        self.invert = g['INVERT']
        self.chord = g['CHORD']
        self.artic = g['ARTIC']
        self.seqRnd = g['SEQRND']
        self.seqRndWeight = g['SEQRNDWT']
        self.compress = g['COMPRESS']
        self.chordRange = g['RANGE']
        self.dupRoot = g['DUPROOT']
        self.chordLimit = g['LIMIT']
        self.unify = g['UNIFY']
        self.midiClear = g['MIDICLEAR']
        self.midiSeq = g['MIDISEQ']
        self.midiVoice = g['MIDIVOICE']
        self.spanStart, self.spanEnd = g['SPAN']
        self.mallet, self.malletDecay = g['MALLET']
        self.ornaments = g['ORNAMENT']

        # It's quite possible that the track was created after
        #  the groove was saved. This means that the data restored
        #  was just the default stuff inserted when the track
        #  was created ... which is fine, but the sequence size
        #  isn't necs. right. We can probably test any list, and octave[]
        #  is as good as any.

        if len(self.octave) != gbl.seqSize:
            self.setSeqSize()

    ####################################
    ## Sequence functions

    def clearSequence(self):
        """ Clear sequence for track.

            This is also called from __init__() to set the initial defaults for each track.

        """

        if self.vtype != 'SOLO' or not self.inited:
            if self.vtype == 'PLECTRUM':
                self.artic = [0]
            else:
                self.artic = [90]
            self.chord = [[]]
            self.sequence = [None]
            self.seqRnd = 0
            self.seqRndWeight = [1]
            self.scaleType = ['AUTO']
            self.rVolume = [[0, 0]]
            self.rPitch = None
            self.rSkip = [0]
            self.rSkipBeats = []
            self.rTime = [[0, 0]]
            self.rDuration = [[0, 0]]
            self.octave = [4 * 12]
            if self.vtype == 'DRUM':
                self.voice = [defaultDrum]
            else:
                self.voice = [defaultVoice]
            self.chordRange = [1]
            self.harmony = [None]
            self.harmonyOnly = [None]
            self.harmonyVolume = [.8]
            self.strum = [None]
            self.strumAdd = [0]
            self.volume = [mma.MMA.volume.vols['M']]
            self.compress = [0]
            self.dupRoot = [[]]
            self.chordLimit = 0
            self.invert = [[]]
            self.lastChord = []
            self.accent = [[]]
            self.unify = [0]
            self.midiClear = []
            self.midiSeq = []
            self.midiVoice = []
            self.spanStart = 0
            self.spanEnd = 127
            self.trigger = mma.MMA.trigger.Trigger()
            self.mallet = 0
            self.malletDecay = 0
            self.futureVols = []
            self.direction = ['BOTH']
            self.delay = [0]

            # ornamentation
            self.ornaments = mma.MMA.ornament.default()

            # for midinote (see midinote.py)
            self.transpose = 0   # transpose off=0, on=1
            self.useticks = 1   # offsets   beats=0, ticks=1  defaults to ticks
            self.tickdur = 1   # duration  notes=0, ticks=1  defaults to ticks
            self.articulate = 0   # honor articulate, defaults to Off
            self.tadjust = 0   # time adjustment factor
            self.vadjust = 1   # volume adjustment factor (percentage)
            self.oadjust = 0   # octave (12 pitches) adjustment

        if self.riff:
            if len(self.riff) > 1:
                warning("%s sequence clear deleting %s riffs" % (self.name, len(self.riff)))
            else:
                warning("%s sequence clear deleting unused riff" % self.name)

        self.riff = []

        self.setSeqSize()

    ############################
    ### Pattern functions
    ############################

    def definePattern(self, name, ln):
        """ Define a Pattern.

        All patterns are stored in pats{}. The keys for this
        are tuples -- (track type, pattern name).

        """

        name = name.upper()
        slot = (self.vtype, name)

        # This is just for the debug code

        if name.startswith('_'):
            redef = "dynamic define"
        elif slot in pats:
            redef = name + ' redefined'
        else:
            redef = name + ' created'

        if not self.vtype in ('SOLO', 'MELODY'):
            ln = ln.rstrip('; ')    # Delete optional trailing    ';' & WS

        pats[slot] = self.defPatRiff(ln)

        if mma.MMA.debug.pshow:
            dPrint("%s pattern %s: %s" % (self.name.title(), redef,
                  self.formatPattern(pats[slot])))

    def dupRiff(self, ln):
        """ Duplicate an existing set of riffs from one solo track to another."""

        # Try to copy From another track to here.

        if ln and ln[0].upper() == 'FROM':
            if len(ln) != 2:
                error("%s DupRiff: FROM option needs exactly one track to copy from." 
                      % self.name)

            t = ln[1].upper()
            if not t in gbl.tnames:
                error("%s DupRiff: Source track '%s' does not exist." % (self.name, t))
            
            tr = gbl.tnames[t]
            
            if self.vtype != tr.vtype:
                error("%s DupRiff: Can't copy from %s, incompatible types (%s != %s)."
                          % (self.name, t, self.vtype, tr.vtype))

            if self.riff:
                error("%s DupRiff: This track had pending data." % self.name)

            if not tr.riff:
                error("%s DupRiff: Source track '%s' has no data to copy." %
                      (self.name, t))

            self.riff = copy.deepcopy(tr.riff)

            if mma.MMA.debug.debug:
                dPrint("%s DupRiff copied from %s." % (self.name, t))

        else:
            if ln and ln[0].upper() == 'TO':  # Optional keyword
                ln = ln[1:]

            if not len(ln):
                error("%s DupRiff: No destination track specified.")

            if not self.riff:
                error("%s DupRiff: No data to copy." % self.name)

            for t in ln:
                t = t.upper()

                if not t in gbl.tnames:
                    error("%s DupRiff: Destination track %s does not exist."
                          % (self.name, t))

                tr = gbl.tnames[t]

                if self.vtype != tr.vtype:
                    error("%s DupRiff: Can't copy to %s, incompatible types (%s != %s)."
                          % (self.name, t, self.vtype, tr.vtype))

                if tr.riff:
                    error("%s DupRiff: Destination track %s has pending data."
                          % (self.name, tr.name))

                tr.riff = copy.deepcopy(self.riff)

                if mma.MMA.debug.debug:
                    dPrint("%s DupRiff copied to %s." % (self.name, tr.name))

    def setRiff(self, ln):
        """ Define and set a Riff. """

        solo = self.vtype in ("MELODY", "SOLO")

        if solo:
            self.riff.append(ln)
        else:
            ln = ln.rstrip('; ')
            if len(ln) == 1 and (ln[0] in ('Z', 'z', '-')):
                self.riff.append([])
            else:
                self.riff.append(self.defPatRiff(ln))

        if mma.MMA.debug.pshow:
            msg = ["%s Riff:" % self.name]
            if solo:
                msg.append(self.riff[-1])
            else:
                msg.append(self.formatPattern(self.riff[-1]))
            dPrint(' '.join(msg))


    def defPatRiff(self, ln):
        """ Worker function to define pattern. Shared by definePattern()
        and setRiff().
        """
        
        def mulPatRiff(oldpat, fact):
            """ Multiply a pattern. """

            fact = stoi(fact, "The multiplier arg must be an integer not '%s'" % fact)

            if fact < 1 or fact > 100:
                error("The multiplier arg must be in the range 2 to 100")

            """ Make N copies of pattern, adjusted so that the new copy has
                all note lengths and start times  adjusted.
                  eg: [[1, 2, 66], [3, 2, 88]]  * 2
                  becomes [[1,4,66], [2,4,88], [3,4,66], [4,4,88]].
            """

            new = []
            add = 0
            step = gbl.barLen // fact

            for n in range(fact):
                orig = copy.deepcopy(oldpat)
                for z in orig:
                    z.offset = (z.offset // fact) + add
                    z.duration //= fact
                    if z.duration < 1:
                        z.duration = 1

                    new.append(z)
                add += step

            return tuple(new)

        def shiftPatRiff(oldpat, fact):

            fact = stof(fact, "The shift arg must be a value, not '%s'" % fact)

            # Adjust all the beat offsets

            new = copy.deepcopy(oldpat)
            max = gbl.barLen
            for n in new:
                n.offset += fact * gbl.BperQ
                if n.offset < 0 or n.offset > max:
                    error("Pattern shift with factor %f has resulted in an "
                          "illegal offset" % fact)

            return tuple(new)

        ### Start of main function...

        # Convert the string to list...
        #  "1 2 3; 4 5 6" --->    [ [1,2,3], [4,5,6] ]

        p = []
        ln = ln.upper().split(';')
        for l in ln:
            p.append(l.split())

        plist = []

        for ev in p:
            more = []
            for i, e in enumerate(ev):
                if e in ('SHIFT', '*'):
                    if i == 0:
                        error("Pattern definition can't start with SHIFT or *")
                    more = ev[i:]
                    ev = ev[:i]
                    break

            if len(ev) == 1:
                nm = (self.vtype, ev[0])

                if nm in pats:
                    if nm[0].startswith('_'):
                        error("You can't use a pattern name beginning with an underscore")
                    pt = pats[nm]

                else:
                    error("%s is not an existing %s pattern" % (nm[1], nm[0].title()))

            else:
                pt = [self.getPgroup(ev)]

            while more:
                cmd = more.pop(0)
                if cmd not in ('SHIFT', '*'):
                    error("Expecting SHIFT or *, not '%s'" % cmd)

                if not more:
                    error("Expecting factor after %s" % cmd)
                if cmd == 'SHIFT':
                    pt = shiftPatRiff(pt, more.pop(0))
                elif cmd == '*':
                    pt = mulPatRiff(pt, more.pop(0))

            plist.extend(pt)

        plist.sort( key=lambda plist: plist.offset)

        if mma.MMA.swing.mode:
            plist = mma.MMA.swing.pattern(plist, self.vtype)

        return plist


    def formatPattern(self, pat):
        pp = []
        if not pat:
            return ' z '

        for p in pat:
            s = []
            s.append("%g" % (1 + (p.offset / float(gbl.BperQ))))
            if self.vtype != 'PLECTRUM':
                s.append("%st" % p.duration)

            if self.vtype == 'CHORD':
                for a in p.vol:
                    s.append("%g" % a)

            elif self.vtype == 'PLECTRUM':
                s.append("%s" % p.strum)
                for a in p.vol:
                    s.append("%g" % a)

            elif self.vtype == 'BASS':
                f = str(p.noteoffset + 1)

                if p.accidental == 1:
                    f += "#"
                elif p.accidental == -1:
                    f += "b"

                if p.addoctave > 0:
                    f += "+" * (p.addoctave // 12)
                elif p.addoctave < 0:
                    f += "-" * (p.addoctave // -12)

                s.append("%s %g" % (f, p.vol))

            elif self.vtype in ('ARPEGGIO', 'SCALE', 'DRUM', 'WALK'):
                s.append("%g " % p.vol)

            pp.append(' '.join(s))
        return "; ".join(pp)

    def insertVoice(self, sc):
        """ Called from bar() and setForceOut(). Adds voice stuff to track."""

        """ 1st pass for MIDIVOICE. There's a separate slot for
            each bar in the sequence, plus the data can be sent
            before or after 'voice' commands. This first loop
            sends MIDIVOICE data with an offset of 0. Note, we
            don't set the value for 'self.smidiVoice' until we
            do this again, later. All this is needed since some
            MIDIVOICE commands NEED to be sent BEFORE voice selection,
            and others AFTER.
        """

        if self.midiVoice:
            v = self.midiVoice[sc]
            if v and v != self.smidiVoice:
                for i in v:
                    if not i[0]:
                        gbl.mtrks[self.channel].addCtl(gbl.tickOffset, i[1])

        # Set the voice in the midi track if not previously done.

        v = self.voice[sc]
        if v != self.ssvoice:

            gbl.mtrks[self.channel].addProgChange(gbl.tickOffset, v, self.ssvoice)
            self.ssvoice = v

            # Mark ssvoice also in shared tracks

            for a in gbl.midiAssigns[self.channel]:
                try:
                    gbl.tnames[a].ssvoice = v
                except KeyError:
                    pass

            if mma.MMA.debug.debug:
                dPrint("%s Voice '%s' inserted at %s" % 
                    (self.name, mma.MMA.midiC.valueToInst(v), gbl.tickOffset))

        """ Our 2nd stab at MIDIVOICE. This time any sequences
            with offsets >0 are sent. AND the smidiVoice and midiSent
            variables are set.
        """

        if self.midiVoice:
            v = self.midiVoice[sc]
            if v and v != self.smidiVoice:
                for i in v:
                    if i[0]:
                        gbl.mtrks[self.channel].addCtl(gbl.tickOffset, i[1])
                self.smidiVoice = v
                self.midiSent = 1  # used by MIDICLEAR

    #########################
    ## Music processing
    #########################

    def bar(self, ctable):
        """ Process a bar of music for this track. """

        # Future vol == de(cresc). Done if track is on or off!

        if self.futureVols:
            self.volume = seqBump([self.futureVols.pop(0)])
        if self.futureVols:
            self.nextVolume = self.futureVols[0]
        else:
            self.nextVolume = None

        # If track is off don't do anything else.

        if self.disable:
            if self.riff:
                self.riff.pop(0)
            return

        """ Decide which seq to use. This is either the current
            seqCount, or if SeqRnd has been set for the track
            it is a random pattern in the sequence.

            The class variable self.seq is set to the sequence to use.
        """

        if self.seqRnd:
            self.seq = mma.MMA.seqrnd.getrndseq(self.seqRndWeight)
        else:
            self.seq = gbl.seqCount

        sc = self.seq

        # Get pattern for this sequence. Either a Riff, Pattern or trigger
        if self.riff:
            pattern = self.riff.pop(0)
        else:
            pattern = self.sequence[sc]
            if self.trigger.mode:
                pattern = mma.MMA.trigger.makeTriggerSequence(self, ctable, pattern)
                if not pattern and self.trigger.override:
                    pattern = self.sequence[sc]
            if not pattern:
                return

        """ MIDI Channel assignment. If no channel is assigned try
            to find an unused number and assign that.
        """

        if not self.channel:
            self.setChannel()

        mma.MMA.midifuncs.doChannelInit(self.channel, self.name)

        # If there is track specific chord data, substitute it for the ctable
        # data already passed.

        if self.chord[sc]:
            ctable = mma.MMA.parse.parseChordLine(self.chord[sc])

        # We are ready to create musical data. 1st do pending midi commands.
        # It probably doesn't matter if we do this now or later??? 

        self.clearPending()

        # This absolutely needs to be done BEFORE any note ON stuff!
        self.insertVoice(sc)

        # Do MIDISeq for this voice

        if self.midiSeq:
            l = self.midiSeq[sc]
            if l:
                for i in l:
                    gbl.mtrks[self.channel].addCtl(getOffset(i[0]), i[1])
                self.midiSent = 1

        """ Special for TRUNCATE. If bar(s) are being truncated patterns
            need to be resized to fit the new size. (Note: SOLO patterns
            are not touched since they are not yet compiled. They will error 
            if they are too long!!)
            We use the 'side' and 'length' settings to figure out where
            to start/end slicing the pattern.
            NOTE: use a deepcopy of the pattern!
        """

        if mma.MMA.truncate.length and self.vtype not in ("SOLO", "MELODY"):
            new = []
            pstart = mma.MMA.truncate.side
            pend = pstart + mma.MMA.truncate.length

            # Go though original pattern and duplicate events (with
            # adjusted offsets) for the new beats. Note that the size/len
            # of pattern is NOT related to the seqsize!

            for a in copy.deepcopy(pattern):
                if a.offset >= pstart and a.offset < pend:
                    a.offset -= pstart
                    new.append(a)

            pattern = new # replace existing pattern with newly created one.

        if pattern:
            self.trackBar(pattern, ctable)

    def clearPending(self):
        self.midiPending.sort()  # probably not needed, but it's cheap insurance
        while self.midiPending:
            c, off, v = self.midiPending.pop(0)

            if c == 'TNAME':
                gbl.mtrks[self.channel].addTrkName(off, v)
                if mma.MMA.debug.debug:
                    dPrint("%s Track name inserted at offset %s" % 
                          (self.name, off))

            elif c == 'GLIS':
                gbl.mtrks[self.channel].addGlis(off, v)
                if mma.MMA.debug.debug:
                    dPrint("%s Glis at offset %s set to %s" % 
                          (self.name, off, ord(chr(v))))

            elif c == 'PAN':
                gbl.mtrks[self.channel].addPan(off, v)
                if mma.MMA.debug.debug:
                    dPrint("%s Pan at offset %s set to %s" % 
                          (self.name, off, v))

            elif c == 'CVOLUME':
                gbl.mtrks[self.channel].addChannelVol(off, v)
                if mma.MMA.debug.debug:
                    dPrint("%s ChannelVolume at offset %s set to %s" % 
                          (self.name, off, v))

            elif c == 'MIDITEXT':
                gbl.mtrks[self.channel].addText(off, v)
                if mma.MMA.debug.debug:
                    dPrint("%s MidiText inserted at %s." % (self.name, off))

            elif c == 'MIDICUE':
                gbl.mtrks[self.channel].addCuePoint(off, v)
                if mma.MMA.debug.debug:
                    dPrint("%s MidiCue inserted at %s." % (self.name, off))

            elif c == 'WHEEL':
                gbl.mtrks[self.channel].addWheel(off, v)
                if mma.MMA.debug.debug:
                    dPrint("%s Wheel inserted at %s." % (self.name, off))

            else:
                error("Unknown midi command pending. Call Bob")

    def getChordInPos(self, offset, ctabs):
        """ Compare an offset to a list of ctables and return
            the table entry active for the given beat.

            The chord start/end offsets generated by parseChordLine() will be in
            the range 0... (BperQ*QperBar). So, negative offset in patterns need
            to be reset to 0; and a hit at the end of the bar could be missed if we
            don't assume that anything out-of-range is in the last chord. Sort of
            ugly, but it's quick and it works.

            Returns a CTable structure.
        """

        for c in ctabs:
            if offset < c.chEnd:
                break
        self.currentChord = c  # for others who might need chord (RPITCH)
        
        return c

    def adjustVolume(self, v, beat):
        """ Adjust a note volume based on the track and global volume
            setting. This is an expensive operation applied to each
            and every generated note.
        """

        # If the volume (velocity) is OFF, just return OFF
        if not v:
            return 0

        sc = self.seq

        # Random skip and random skipbeats, return OFF
        if self.rSkip[sc] and (not self.rSkipBeats or beat in self.rSkipBeats) \
                and random.random() < self.rSkip[sc]:
            return 0

        # If the track is set to OFF, return OFF
        trackAdjust = self.volume[sc]
        if not trackAdjust:
            return 0

        # single bar cresc adjust, adjust track volume adjuster value
        if self.nextVolume: 
            bt = beat
            if bt < 1:  # might have negative offsets, cres code ignores
                bt = 0
            trackAdjust += (self.nextVolume - trackAdjust) * \
                bt / gbl.barLen

        trackAdjust *= mma.MMA.volume.vTRatio

        fullAdjust = mma.MMA.volume.volume

        # Is everything off?
        if not fullAdjust:
            return 0

        # global single bar cresc adjust
        if mma.MMA.volume.nextVolume:
            bt = beat
            if bt < 1:  # might have negative offsets, cres code ignores
                bt = 0
            fullAdjust += (mma.MMA.volume.nextVolume - fullAdjust) * \
                bt / gbl.barLen

        fullAdjust *= mma.MMA.volume.vMRatio

        v *= (trackAdjust + fullAdjust)

        for b, a in self.accent[sc]:
            if b == beat:
                v += (v * a)

        # Random volume adjustment
        if self.rVolume[sc]:
            trackAdjust = int(v * self.rVolume[sc][0])
            fullAdjust = int(v * self.rVolume[sc][1])
            if trackAdjust or fullAdjust:
                if fullAdjust < trackAdjust:
                    trackAdjust, fullAdjust = fullAdjust, trackAdjust
                v += random.randrange(trackAdjust, fullAdjust)

        # normalize
        if v > 127:
            v = 127
        elif v < 1:
            v = 1

        # all was done in fp, don't forget to round!
        return int(v)

    def adjustNote(self, n):
        """ Adjust a note for a given octave/transposition.
            Ensure that the note is in range.
        """

        n += self.octave[self.seq] + gbl.transpose

        while n < self.spanStart:  # spanStart will be 0 from init
            n += 12
        while n > self.spanEnd:  # spanEnd will be 127 from init
            n -= 12

        return n


    def setBarOffset(self, v, emsg=''):
        """ Convert a string into a valid bar offset in midi ticks. """

        m = v.find('-')
        p = v.find('+')

        # determine split point and sign
        if m > -1 and p > -1:  # both - and +, get lowest
            if m > p:
                sp = p
                sign = 1
            else:
                sp = m
                sign = -1

        elif m > -1:   # only -, easy
            sp = m
            sign = -1

        elif p > -1:  # only +, easy
            sp = p
            sign = 1

        else:
            sp = None  # easier, no split

        if sp:   # get offset, and note length modifier
            note = v[sp + 1:]
            v = v[:sp]
            notepart = mma.MMA.notelen.getNoteLen(note) * sign
        else:
            notepart = 0

        v = stof(v, "%sValue for %s bar offset must be integer/float, not '%s'" %
                 (emsg, self.name, v))
        v = ((v - 1) * gbl.BperQ) + notepart

        if v < 0:
            if v < -gbl.BperQ:
                error("%sValue for %s bar offset must be 0 or greater, not '%s'" %
                      (emsg, self.name, v/gbl.BperQ+1))
            else:
                warning("Offset in '%s' is '%s ticks' before bar start!" % (self.name, -v))

        if v >= gbl.barLen:
            error("%sValue for %s bar offset must be less than %s, not '%s'." %
                  (emsg, self.name, gbl.QperBar + 1, v/gbl.BperQ+1))

        return int(v)

    def getDur(self, d):
        """ Return the adjusted duration for a note.

            Articulate and RDuration are used.
        """

        sc = self.seq

        d = (d * self.artic[sc]) // 100

        if self.rDuration[sc]:
            a1 = int(d * self.rDuration[sc][0])
            a2 = int(d * self.rDuration[sc][1])
            if a1 or a2:
                d += random.randrange(a1, a2)
        if d <= 0:
            d = 1   # force a value if we end with 0 or less.

        return int(d)

    def sendNote(self, offset, duration, note, velocity):
        """ Send a note to the MIDI machine. This is called from all
            track classes and handles niceties like mallet-repeat, rpitch
            and delay.
        """

        if not velocity:
            return

        sc = self.seq

        if self.rPitch:
            note = mma.MMA.rpitch.doRpitch(self, note)
        rptr = self.mallet
        if rptr and duration > rptr:
            ll = self.getDur(rptr)
            offs = 0
            vel = velocity
            for q in range(duration // rptr):
                gbl.mtrks[self.channel].addPairToTrack(
                    offset + offs + self.delay[sc],
                    self.rTime[sc][0], self.rTime[sc][1],
                    ll,
                    note,
                    vel,
                    None)

                offs += rptr
                if self.malletDecay:
                    vel = int(vel + (vel * self.malletDecay))
                    if vel < 1:
                        vel = 1
                    if vel > 127:
                        vel = 127

        else:
            gbl.mtrks[self.channel].addPairToTrack(
                offset + self.delay[sc],
                self.rTime[sc][0], self.rTime[sc][1],
                duration,
                note,
                velocity,
                self.unify[sc])

    def sendChord(self, chord, duration, offset):
        """ Send a chord to midi with strum delay. Called
            from Chord, Bass, Walk, Arpeggio and Scale.

            chord = list of [note, volume]s
            p = pattern for current beat

            This compensates out the strum value so that the additional (harmony)
            notes all end at the same time.

            Adjusts the RDURATION setting.
        """

        sc = self.seq
        strumAdjust = self.getStrum(sc)
        strumAdd = self.strumAdd[sc]

        if strumAdjust:
            minDuration = duration // 3
            if minDuration < gbl.Bper128:   # this is 12 ticks
                minDuraiton = gbl.Bper128

        strumOffset = 0  # 1st note in list is not offset-strummed

        duration = self.getDur(duration)
        for n, vol in chord:
            self.sendNote(
                int(offset + strumOffset),
                duration,
                self.adjustNote(n),
                self.adjustVolume(vol, offset))

            if strumAdjust:
                strumAdjust += strumAdd
                strumOffset += strumAdjust
                duration -= strumAdjust
                if duration < minDuration:
                    duration = minDuration



# globals.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

version = "20.02"        # Version -- Feb/2020

""" A few globals are actually set in the calling stub, mma.MMA.py. This is
    done to make future ports and platform specific settings a bit easier.
    The following variables are imported from mma.MMA.py and stored here:

        platform   - host platform, Windows, Linux, etc.
        MMAdir     - the home directory for mma stuff

    The above variables can be accessed from the rest of the mma modules in
    the form "gbl.MMAdir", etc.
"""

# from __main__ import MMAdir, platform
from mma import MMAdir, platform_so
platform = platform_so

""" mtrks is storage for the MIDI data as it is created.
    It is a dict of class Mtrk() instances. Keys are the
    midi channel numbers. Ie, mtrks[2]    is for channel 2,
    etc. mtrks[0] is for the meta stuff.
"""

mtrks = {}

""" tnames is a dict of assigned track names. The keys are
    the track names; each entry is a pattern class instance.
    We have tnames['BASS-FOO'], etc.
"""

tnames = {}

""" midiAssigns keeps track of channel/track assignments. The keys
    are midi channels (1..16), the data is a list of tracks assigned
    to each channel. The tracks are only added, not deleted. Right
    now this is only used in -c reporting.
"""

midiAssigns = {}
for c in range(0, 17):
    midiAssigns[c] = []

""" midiAvail is a list with each entry representing a MIDI channel.
    As channels are allocated/deallocated the appropriated slot
    is inc/decremented.
"""

midiAvail = [0] * 17   # slots 0..16, slot 0 is not used.

deletedTracks = []    # list of deleted tracks for -c report

""" This is a user constructed list of names/channels. The keys
    are names, data is a channel. Eg. midiChPrefs['BASS-SUS']==9
"""

midiChPrefs = {}


""" If the -T option is used only the tracks in this list
    are generated. All other tracks are muted (OFF)
"""

muteTracks = []


############# String constants ####################


EXT = ".mma"        # extension for song/lib files (STATC).


##############  Tempo, and other midi positioning.  #############


BperQ       =  192    # midi ticks per quarter note (STATIC)
Bper128     =  BperQ/16  # a 1/128 note. Used for small timings (STATIC)
QperBar     =  4      # Beats/bar, set with TIME (this is a fp value!)
barLen      =  int(BperQ * QperBar)  # convenience (updated by TIME)
tickOffset  =  0      # offset of current bar in ticks
tempo       =  120    # current tempo
seqSize     =  1      # variation sequence table size
seqCount    =  0      # running count of variation

totTime     = 0.0     # running duration count in seconds

transpose   =  0      # Transpose is global (ignored by drum tracks)

lineno      = -1      # used for error reporting

barLabels   = []      # a list of bar (number) labels as encountered
barNum      =  0      # Current line number

barPtrs     = {}      # for each bar, pointers to event start/end

outPath    =   ''      # Directory for MIDI file
inpath     =   None    # input file

midiFileType   = 1     # type 1 file, "MidiFile SMF" command can change to 0/1
runningStatus  = 1     # running status enabled "MidiFile Running" changes to 0/1

inAllGrooves = False   # set if running an ALLGROOVES command

encoding = 'cp1252'
noCredit = False   # set to not have credits in midi file
logFile = ''    # this is set if user wants logging to be buffered.

#############  Options. #############


""" These variables are all set from the command line in mma.MMA.opts.py.
    It's a bit of an easy-way-out to have them all here, but I don't think
    it hurts too much.
"""

barRange       =     []      # both -B and -b use this

printProcessed = False  # command line flag -L sets this

outfile        =     None
infile         =     None
createDocs     =     0
maxBars        =     500
makeGrvDefs    =     0

playFile       =     0       # set if we want to call a player

# set if an exception is catching unknown chords
# this is used by -xCHORDS 
ignoreBadChords = False





# grooves.py


"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""


""" Groove storage. Each entry in glist{} has a keyname
    of a saved groove.

    lastGroove and currentGroove are used by macros
"""

import mma.MMA.midi
import mma.MMA.notelen
import mma.MMA.auto
import mma.MMA.volume
import mma.MMA.parse
import mma.MMA.parseCL
import mma.MMA.seqrnd
import mma.MMA.docs
import mma.MMA.debug

from mma.MMA.timesig import timeSig
from . import gbl
from   mma.MMA.common import *


glist = {}
aliaslist = {}

lastGroove = ''     # name of the last groove (used by macros)
currentGroove = ''     # name of current groove (used by macros)

COPYGROOVE = -2387      # fake groove name used by copytrack

""" groovesList[] holds a list of grooves to use when the
     groove command is used with several args. ie, when
     we use either:

         Groove 2  groove1  groove2 groove3
     or
         Groove groove1 groove2 groove3

     in both cases, the names of the grooves are stored in groovesList[]
"""

groovesList = None
groovesCount = 0


class STACKGROOVE():
    """ A little stacker for temporary grooves. """

    def __init__(self):
        self.stack = []   # set of integers used as groove names

    def push(self):
        s = self.stack
        s.append(len(s))
        grooveDefineDo(s[-1])

    def pop(self):
        s = self.stack
        if not s:
            error("Groove stacking error. Call Bob.")
        grooveDo(s.pop())

stackGroove = STACKGROOVE()   # single (only) instance


def grooveDefine(ln):
    """ Define a groove.

        Current settings are assigned to a groove name.
    """

    if not len(ln):
        error("Use: DefGroove  Name")

    slot = ln[0].upper()

    # Slot names can't contain a '/' or ':'

    if '/' in slot:
        error("The '/' is not permitted in a groove name.")

    if ':' in slot:
        error("The ':' is not permitted in a groove name.")

    if slot in aliaslist:
        error("Can't define groove name %s, already defined as an alias for %s." \
                  % (slot, aliaslist[slot]))

    if mma.MMA.debug.gvShow and slot in glist:
        dPrint("Redefining groove %s, line %s." % (slot, gbl.lineno))

    grooveDefineDo(slot)

    if mma.MMA.debug.debug:
        dPrint("Groove settings saved to '%s'." % slot)

    if gbl.makeGrvDefs:   # doing a database update ...
        mma.MMA.auto.updateGrooveList(slot)

    if len(ln) > 1:
        mma.MMA.docs.docDefine(ln)


def grooveDefineDo(slot):

    for n in gbl.tnames.values():
        if n.sticky:
            continue
        n.saveGroove(slot)

    glist[slot] = {
        'SEQSIZE':   gbl.seqSize,
        'SEQRNDWT':  mma.MMA.seqrnd.seqRndWeight[:],
        'QPERBAR':   gbl.QperBar,
        'BARLEN':    gbl.barLen,
        'SEQRND':    mma.MMA.seqrnd.seqRnd[:],
        'TIMESIG':   timeSig.get(),
        'SWINGMODE': mma.MMA.swing.gsettings(),
        'VRATIO':    (mma.MMA.volume.vTRatio, mma.MMA.volume.vMRatio),
        'CTABS':     mma.MMA.parseCL.chordTabs[:] }


def grooveAlias(ln):
    """ Create an alias name for an existing groove. Note: not a copy. """

    global aliaslist

    if len(ln) != 2:
        error("DefAlias needs exactly 2 args: GrooveName AliasName.")

    g = ln[0].upper()
    a = ln[1].upper()

    if not g in glist:
        error("DefAlias: Groove %s has not been defined." % ln[0])

    aliaslist[a] = g


def groove(ln):
    """ Select a previously defined groove. """

    global groovesList, groovesCount, lastGroove, currentGroove

    if not ln:
        error("Groove: needs agrument(s)")

    tmpList = []

    if ln[0].isdigit():
        wh = stoi(ln[0])
        if wh < 1:
            error("Groove selection must be > 0, not '%s'" % wh)
        ln = ln[1:]
    else:
        wh = None

    for slot in ln:
        slotOrig = slot   # we need this for search
        slot = slot.upper()

        if slot == "/":
            if len(tmpList):
                slot = tmpList[-1]
            else:
                error("A previous groove name is needed before a '/'")

        # convert alias to real groove name
        if not slot in glist and slot in aliaslist:
                slot = aliaslist[slot]
        
        if not slot in glist:
            if mma.MMA.debug.debug:
                dPrint("Groove '%s' not defined. Trying auto-load from libraries" 
                      % slot)

            l, slot = mma.MMA.auto.findGroove(slotOrig)    # name of the lib file with groove
            
            if l:
                if mma.MMA.debug.debug:
                    dPrint("Attempting to load groove '%s' from '%s'." % (slot, l))

                reportFutureVols()
                mma.MMA.parse.usefile([l])

                if not slot in glist:
                    error("Groove '%s' not found. Is the groove in the file '%s'? "
                          "Have libraries changed "
                          "since last 'mma -g' run?" % (slot, l))

            else:
                error("Groove '%s' could not be found in memory or library files" % slot )

        tmpList.append(slot)

    if not len(tmpList):
        error("Use: Groove [selection] Name [...]")

    ## If the first arg to list was an int() (ie: 3 groove1 groove2 grooveFoo)
    ## we select from the list. After the selection, we reset the list to be
    ## just the selected entry. This was, if there are multiple groove names
    ## without a leading int() we process the list as groove list changing
    ## with each bar.

    if wh:
        wh = (wh-1) % len(tmpList)
        tmpList = tmpList[wh:wh+1]

    slot = tmpList[0]
    grooveDo(slot)

    groovesCount = 0
    if len(tmpList) == 1:
        groovesList = None
    else:
        groovesList = tmpList

    lastGroove = currentGroove
    currentGroove = slot
    if lastGroove == '':
        lastGroove = slot

    if mma.MMA.debug.debug:
        dPrint("Groove settings restored from '%s'." % slot)


def grooveDo(slot):
    """ This is separate from groove() which parses the command line,
        does a lib search, etc. This just copies data from an existing
        in-memory groove.
    """

    reportFutureVols()

    oldSeqSize = gbl.seqSize

    g = glist[slot]

    gbl.seqSize = g['SEQSIZE']
    mma.MMA.seqrnd.seqRndWeight = g['SEQRNDWT']
    gbl.QperBar = g['QPERBAR']
    gbl.barLen  = g['BARLEN']
    mma.MMA.seqrnd.seqRnd = g['SEQRND']
    timeSig.create(*g['TIMESIG'])  # passing tuple as 2 args.
    mma.MMA.swing.grestore(g['SWINGMODE'])
    mma.MMA.volume.vTRatio, mma.MMA.volume.vMRatio = g['VRATIO']
    mma.MMA.parseCL.chordTabs = g['CTABS']

    for n in gbl.tnames.values():
        if n.sticky:
            continue
        n.restoreGroove(slot)

    ## This is important! Tracks NOT overwritten by saved grooves may
    ## have the wrong sequence length. I don't see any easy way to hit
    ## just the unchanged/unrestored tracks so we do them all.
    ## Only done if a change in seqsize ... doesn't take long to be safe.

    if oldSeqSize != gbl.seqSize:
        for a in gbl.tnames.values():
            a.setSeqSize()

    mma.MMA.seqrnd.seqRndWeight = seqBump(mma.MMA.seqrnd.seqRndWeight)

    gbl.seqCount = 0


def reportFutureVols():
    """ Print warning for pending track cresendos.

        We need a seperate func here since the groove() may
        parse a new file, which will clear out data before
        getting to grooveDo().

        Note that the test is for more that one trailing future volume.
        This is deliberate ... a construct like:

           Chord Cresc ff 1
           ..somechord
           Groove NEW

        will leave a future volume on the stack.
    """

    volerrs = []
    for n in gbl.tnames.values():
        if n.vtype in ("SOLO", "ARIA"):  # not saved in grooves
            continue
        if len(n.futureVols) > 1:
            volerrs.append(n.name)
        n.futureVols = []     # don't want leftover future vols a track level!

    if volerrs:
        volerrs.sort()
        warning("Pending (de)Cresc in %s." % ', '.join(volerrs))


def grooveClear(ln):
    """ Delete all previously loaded grooves from memory. """

    global groovesList, groovesCount, glist, lastGroove, currentGroove, aliaslist

    if ln:
        error("GrooveClear does not have any arguments.")

    groovesList = {}
    aliaslist = {}
    groovesCount = 0
    tmplist = {}

    # we first save all the temp grooves (for include, etc) into a glist replacement

    for a in glist:
        if isinstance(a, int):     # is groove name is an integer, not string
            tmplist[a] = glist[a]  # yes, stacked groove, save

    # now just point glist to the copy. Normally created grooves are gone.

    glist = tmplist

    lastGroove = ''
    currentGroove = ''

    if mma.MMA.debug.debug:
        dPrint("All grooves deleted.")


def nextGroove():
    """ Handle groove lists. Called from parse().

        If there is more than 1 entry in the groove list,
        advance (circle). We don't have to qualify grooves
        since they were verified when this list was created.
        groovesList is None if there is only one groove (or none).
    """

    global lastGroove, currentGroove, groovesCount

    if groovesList:
        groovesCount += 1
        if groovesCount > len(groovesList)-1:
            groovesCount = 0
        slot = groovesList[groovesCount]

        if slot != currentGroove:
            grooveDo(slot)

            lastGroove = currentGroove
            currentGroove = slot

            if mma.MMA.debug.debug:
                dPrint("Groove (list) setting restored from '%s'." % slot)


def trackGroove(name, ln):
    """ Select a previously defined groove for a single track. """

    if len(ln) != 1:
        error("Use: %s Groove Name" % name)

    slot = ln[0].upper()

    # convert alias to real groove name
    if not slot in glist and slot in aliaslist:
        slot = aliaslist[slot]
            

    if not slot in glist:
        error("Groove '%s' not defined" % slot)

    g = gbl.tnames[name]
    g.restoreGroove(slot)

    if g.sequence == [None] * len(g.sequence):
        warning("'%s' Track Groove has no sequence. Track name error?" % name)

    if mma.MMA.debug.debug:
        dPrint("%s Groove settings restored from '%s'." % (name, slot))


def getAlias(al):
    """ This is used by the library doc printer to get a list aliases. """

    al = al.upper()
    l = []

    for a in aliaslist:
        if aliaslist[a] == al:
            l.append(a.title())
  
    return ', '.join(l)


def allgrooves(ln):
    """ Apply a command to all currently defined grooves. """

    if not ln:
        error("AllGrooves: requires arguments.")

    stackGroove.push()  # save the current groove into a temp slot

    # Disable the warning messages ... use with caution.
    if len(ln)>1 and ln[0].upper() == "NOWARN":
        startNoWarn = mma.MMA.debug.noWarn
        mma.MMA.debug.noWarn = 1
        noWarn = True
        ln = ln[1:]
    else:
        noWarn = False
        
    action = ln[0].upper()   # either a command or a trackname
    if len(ln) > 1:
        trAction = ln[1].upper()
    else:
        trAction = ''

    sfuncs = mma.MMA.parse.simpleFuncs
    tfuncs = mma.MMA.parse.trackFuncs
    seqwarning = 0

    counter = 0

    # Set warning level to off. This is done since simple things like:
    #     SomeTrack  Volume m mf mp p
    # will spew error messages because all the grooves in memory might
    # have different seqsize values.

    gbl.inAllGrooves = True

    for g in glist:   # do command for each groove in memory
        grooveDo(g)   # activate the groove

        if action in sfuncs:        # test for non-track command and exe.
            if action in ('GROOVE', 'DEFGROOVE', 'ALLGROOVES', 'ALLTRACKS'
                          'BEATADJUST', 'GROOVECLEAR'):
                error("AllGrooves: '%s' cannot be applied like this." % action)

            sfuncs[action](ln[1:])  # do the command
            counter += 1

        else:                       # see if this is a track command
            if not trAction:
                error("AllGrooves: No command for assumed trackname %s." % action)

            name = action  # remember 'action' is ln[0]. Using 'name' just makes it clearer

            if not name in gbl.tnames:  # skip command if track doesn't exist
                continue

            if trAction in tfuncs:
                if trAction == 'SEQUENCE' and not seqwarning:
                    warning("AllGrooves: Setting SEQUENCE on all grooves is probably a bad idea.")
                    seqwarning = 1   # we only print the warning once

                tfuncs[trAction](name, ln[2:])
                counter += 1

            else:
                error ("AllGrooves: Not a command: '%s'" % ' '.join(ln))

        grooveDefineDo(g)       # store the change!!!

    stackGroove.pop()       # restore original state

    gbl.inAllGrooves = False

    if noWarn:
        mma.MMA.debug.noWarn = startNoWarn
        
    if not counter:
        warning("No tracks affected with '%s'" % ' '.join(ln))

    else:
        if mma.MMA.debug.debug:
            dPrint("AllGrooves: %s tracks modified." % counter)


###################################################################

def trackCopy(name, ln):
    """ Copy/Duplicate all track data to/from existing track. """

    if ln and ln[0].upper() == 'TO':    # arg flipper
        src = name

        for dest in ln[1:]:
            dest = dest.upper()  # Destination is 1st arg
            mma.MMA.alloc.trackAlloc(dest, 1) # create dest if it doesn't exist
            trackCopyDo(dest, [src])

    elif ln and ln[0].upper() == 'FROM':   # optional 'from' keyword
        ln = ln[1:]
        if len(ln) != 1:
            error("Copy FROM %s: Exactly one arg needed (track to copy)." % name)
        trackCopyDo(name, ln)

    else:
        trackCopyDo(name, ln)
        
def trackCopyDo(name, ln):
    """ Copy/Duplicate all track data from existing track.  """

    if len(ln) != 1:
        error("Copy %s: Exactly one arg needed (track to copy)." % name)
    cp = ln[0].upper()

    self = gbl.tnames[name]

    # Split the track-to-copy into a trackname and groovename

    if '::' in cp:
        gr, cp = cp.split('::', 1)
    else:
        gr = ''

    # Find/load the needed groove.
    if gr:
        otime = gbl.QperBar
        stackGroove.push()
        groove([gr.split(':', 1)[0]])              
        if otime != gbl.QperBar:
            error("Copy %s: TIME mismatch, groove %s has time of %s, not %s." %
                  (name, gr, gbl.QperBar, otime))
        
    if not cp in gbl.tnames:
        error("Copy %s: Track '%s' is not defined." % (name, cp))

    cp = gbl.tnames[cp]   # point to the track object

    if cp.vtype != self.vtype:
        error("Copy %s: tracks must be of same type, not %s and %s."
              % (name, self.vtype, cp.vtype))

    # Push the track to copy into a buffer. Use saveGroove for this.
    # Note that solo tracks have a special function since they normally
    # do note save groove info.
    if cp.vtype == 'SOLO':
        if gr:
            error("Copy %s: Solo tracks cannot be copied from a groove, only from memory."
                  % name)
        cp.forceSaveGroove(COPYGROOVE)
    else:
        cp.saveGroove(COPYGROOVE)

    # We need this indirection. The restore function takes
    # a groove buffer name, not a buffer. So it's either rewrite
    # the function or create a scratch name which makes it happy.
    # And it's cheap to do ... no data is copied here.
    self.grooves[COPYGROOVE] = cp.grooves[COPYGROOVE]

    # If we loaded a groove, restore to the original state.
    if gr:
        stackGroove.pop()

    # Now copy the buffered data into the target track.
    if cp.vtype == 'SOLO':
        self.forceRestoreGroove(COPYGROOVE)
    else:
        self.restoreGroove(COPYGROOVE)

    if mma.MMA.debug.debug:
        dPrint("Copy: Settings duplicated from %s to %s" % (cp.name, self.name))



# tweaks.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

This module has support for some simple default tweaks.

"""

from mma.MMA.common import *
import mma.MMA.pat
import mma.MMA.midiC

import copy


def setTweak(ln):
    """ Option tweaks. """

    notopt, ln = opt2pair(ln)

    if notopt:
        error("Tweaks: expecting cmd=opt pairs, not '%s'." % ' '.join(notopt))

    for cmd, opt in ln:
        cmd = cmd.upper()

        if cmd in ('DEFAULTDRUM', 'DEFAULTTONE'):
            mma.MMA.pat.defaultDrum = mma.MMA.midiC.decodeVoice(opt)

        elif cmd == 'DEFAULTVOICE':
            mma.MMA.pat.defaultVoice = mma.MMA.midiC.decodeVoice(opt)

        elif cmd == 'DIM':
            from mma.MMA.chordtable import chordlist

            if opt == '3':
                # this is so we can change the desc. (no non-standard)
                chordlist['dim'] = (chordlist['dim3'][0], chordlist['dim3'][1], "Diminished triad")
            elif opt == '7':
                chordlist['dim'] = copy.deepcopy(chordlist['dim7'])
            else:
                error("Tweaks: DIM requires '3' or '7' arg, not '%s'." % opt)

        else:
            error("Tweaks: '%s' unknown command." % cmd)



# translate.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>


This module handles voice name translations.

"""

import mma.MMA.midiC

from . import gbl
from mma.MMA.common import *
import mma.MMA.debug

""" Translation table for VOICE. This is ONLY used when a voice is set
    from the VOICE command. If a translation exists the translation is
    substituted.
"""


class Vtable:

    def __init__(self):
        self.table = {}

    def retlist(self):

        l = []
        for n in sorted(self.table.keys()):
            l.append("%s=%s" % (n.title(), self.table[n]))

        return ' '.join(l)

    def create(self, ln):
        """ Set a name/alias for voice translation, called from parser. """

        if not ln:
            self.table = {}
            if mma.MMA.debug.debug:
                dPrint("Voice Translation table reset.")
            return

        ln, opts = opt2pair(ln, toupper=True)

        if ln:
            error("VOICETR: Each translation pair must be in the format Alias=Voice.")

        for v, a in opts:
            self.table[v] = a

        if mma.MMA.debug.debug:
            dPrint("Voice Translations: %s" % ' '.join(["%s=%s" % (v, a) for v, a in opts]))

    def get(self, name):
        """ Return a translation or original. """

        name = name.upper()
        try:
            return self.table[name]
        except KeyError:
            return name

vtable = Vtable()            # Create single class instance.


""" This is just like the Vtable, but it is used for DRUM TONES. We use
    this translation when a TONE is set for a drum in setTone() and when a
    tone is selected in a Solo/Melody DRUM track.
"""


class Dtable:

    def __init__(self):
        self.table = {}

    def retlist(self):

        l = []
        for n in sorted(self.table.keys()):
            l.append("%s=%s" % (mma.MMA.midiC.valueToDrum(n),
                                mma.MMA.midiC.valueToDrum(self.table[n])))

        return ' '.join(l)

    def set(self, ln):
        """ Set a name/alias for drum tone translation, called from parser. """

        if not ln:
            self.table = {}
            if mma.MMA.debug.debug:
                dPrint("DrumTone Translation table reset.")

            return

        ln, opts = opt2pair(ln, toupper=True)

        if ln:
            error("TONETR: Each translation pair must be in the format Tone=NewTone.")

        for v, a in opts:
            v1 = mma.MMA.midiC.drumToValue(v)
            if v1 < 0:
                error("TONETR: Tone '%s' not defined." % v)

            a1 = mma.MMA.midiC.drumToValue(a)
            if a1 < 0:
                error("TONETR: Tone '%s' not defined." % a)

            self.table[v1] = a1

        if mma.MMA.debug.debug:
            dPrint("TONETR Translations: %s" % 
                  ' '.join(["%s(%s)=%s" % (v, mma.MMA.midiC.drumToValue(v), 
                                           mma.MMA.midiC.drumToValue(a)) for v, a in opts]))


    def get(self, name):
        """ Return a translation or original. Note that this also does
            validation of 'name'. It is called from patDrum and patSolo.
        """

        v = mma.MMA.midiC.drumToValue(name)

        if v < 0:
            error("Drum Tone '%s' not defined." % name)

        try:
            return self.table[v]
        except KeyError:
            return v


dtable = Dtable()

""" Volume adjustment. Again, similar to voice/tone translations,
    but this is for the volume. The table creates a percentage adjustment
    for tones/voices specified. When a TRACK VOLUME is set in
    MMApat.setVolume() the routine checks here for an adjustment.
"""


class VoiceVolTable:

    def __init__(self):
        self.table = {}

    def retlist(self):
        l = []
        for n in sorted(self.table.keys()):
            l.append("%s=%s" % (mma.MMA.midiC.valueToInst(n), self.table[n]))

        return ' '.join(l)

    def set(self, ln):
        """ Set a name/alias for voice volume adjustment, called from parser. """

        if not ln:
            self.table = {}
            if mma.MMA.debug.debug:
                dPrint("Voice Volume Adjustment table reset.")

            return

        ln, opts = opt2pair(ln)

        if ln:
            error("VOICEVOLTR: Expecting VOICE=VOLUME pairs.")

        for v, a in opts:
            val = mma.MMA.midiC.instToValue(v)

            if val < 0:
                error("VOICEVOLTR: unknown voice '%s'." % v)

            a = stoi(a)
            if a < 1 or a > 200:
                error("VOICEVOLTR: adjustments must be in range 1 to 200, not '%s'." % a)

            self.table[val] = a / 100.

        if mma.MMA.debug.debug:
            dPrint("VOICEVOLTR: %s" % ' '.join(["%s=%s" % (v.upper(), a) for v, a in opts]))

    def get(self, v, vol):
        """ Return an adjusted value or original. """

        try:
            return vol * self.table[v]
        except:
            return vol

voiceVolTable = VoiceVolTable()


class DrumVolTable:

    def __init__(self):
        self.table = {}

    def retlist(self):

        l = []
        for n in sorted(self.table.keys()):
            l.append("%s=%s" % (mma.MMA.midiC.valueToDrum(n), self.table[n]))

        return ' '.join(l)

    def set(self, ln):
        """ Set a name/alias for drumtone volume adjustment, called from parser. """

        if not ln:
            self.table = {}
            if mma.MMA.debug.debug:
                dPrint("DRUMVOLTR: Adjustment table reset.")

            return

        ln, opt = opt2pair(ln, toupper=True)

        if ln:
            error("DRUMVOLTR: Each option must be in the format TONE=AJUSTMENT.")

        for v, a in opt:
            val = mma.MMA.midiC.drumToValue(v)

            if val < 0:
                error("DRUMVOLTR: Unknown tone '%s'." % v)

            a = stoi(a)
            if a < 1 or a > 200:
                error("DRUMVOLTR: adjustments must be in range 1 to 200, not '%s'." % a)

            self.table[val] = a / 100.

        if mma.MMA.debug.debug:
            dPrint("DRUMVOLTR: %s" %
                  ' '.join(["%s=%s" % (mma.MMA.midiC.valueToDrum(val), a) for v, a in opt]))


    def get(self, v, vol):
        """ Return an adjusted value or original. """

        try:
            return vol * self.table[v]

        except:         # not the best, but any errors just return the orignal volume.
            return vol

drumVolTable = DrumVolTable()



# truncate.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>


This code just sets up truncate. The heavy lifting is done in pat.py/bar()
where the sequence for each track is re-created with adjusted offsets.
In parse.py/parse() we adjust the song pointer for the current seq size.

"""

from mma.MMA.common import *
import mma.MMA.debug

length = None
count = None
side = None


def setTruncate(ln):
    """ Set the truncate variable for the next bar. """

    global length, count, side

    if length:
        warning("Truncate: option set with value pending, previous setting discarded.")

    side = 0    # assume 'right' or start of bar
    count = 1    # assume 1 bar only

    ln, opts = opt2pair(ln)   # separate out the option strings

    # Grab the truncate length first, need this to figure the
    # value for the side option, later.

    if len(ln) != 1:
        error("Truncate: Beats must be set. Syntax is <beats> [option=value]")

    beats = stof(ln[0], "Truncate: Bar length must be value.")

    if beats < 1 or beats >= gbl.QperBar:
        error("Truncate: Range must be 1 to %s, not '%s'." % (gbl.QperBar, beats))

    length = int(beats * gbl.BperQ)

    # now parse options

    for cmd, opt in opts:
        cmd = cmd.upper()

        if cmd == 'COUNT':
            b = stoi(opt, "Truncate: Bar COUNT must be integer.")
            if b < 1:
                error("Truncate: Bar COUNT must be 1 or greater.")
            count = b

        elif cmd == "SIDE":
            opt = opt.upper()
            max = gbl.barLen

            if opt == "LEFT":   # side to use, default==left
                side = 0

            elif opt == "RIGHT":  # use the right side of pattern
                side = max - length

            else:
                opt = stof(opt, "Truncate: Expecting value, not '%s'." % opt)
                side = int((opt - 1) * gbl.BperQ)
                if side >= gbl.barLen:
                    error("Truncate: Side value of '%g' too large." % opt)

        else:
            error("Truncate: '%s' is an unknown option." % cmd)

    if mma.MMA.debug.debug:
        dPrint("Truncate: Next %s bar(s) are %g beats, "
              "using pattern from beats %g to %g."
            % (count, beats, float(side) / gbl.BperQ, (float(side) + length) / gbl.BperQ))


def countDown():
    """ Decrement the bar count. Called from parse.  """

    global length, count, side

    count -= 1
    if count <= 0:
        count = None
        length = None
        side = None



# patSolo.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

import mma.MMA.notelen
import mma.MMA.translate
import mma.MMA.harmony
import mma.MMA.volume
import mma.MMA.alloc
import mma.MMA.swing
import mma.MMA.truncate

from . import gbl

from mma.MMA.common import *
from mma.MMA.pat import PC, Pgroup
from mma.MMA.keysig import keySig
import mma.MMA.debug

import re
import random


# Each note in a solo gets a NoteEvent.
class NoteEvent:
    def __init__(self, pitch, velocity, isgrace):
        self.duration = None
        self.pitch = pitch
        self.articulation = None
        self.velocity = velocity
        self.defvelocity = velocity
        self.isgrace = isgrace             # signal grace note, no harmony and change offset

accValues = {'#': 1, "&": -1, 'n': 0}

# used when extracting solo notes.
defaultVelocity = 90

##############################


class Melody(PC):
    """ The melody and solo tracks are identical, expect that
        the solo tracks DO NOT get saved in grooves and are only
        initialized once.
    """

    vtype = 'MELODY'
    drumType = None

    endTilde = []
    drumTone = 38
    arpRate = 0
    arpDecay = 0
    arpDirection = 'UP'
    stretch = 1
    followChord = 0
    followKey = 1
    rootChord = 0   # start off with a C chord as default

    def __init__(self, ln):
        """ We need special init here incase this is converted to drumType.
            If we don't have a toneList, we're screwed.
        """

        self.toneList = [self.drumTone]
        PC.__init__(self, ln)   # This order is important!

    def setDrumType(self):
        """ Set this track to be a drum track. """

        if self.channel:
            error("You cannot change a track to DRUM once it has been used")

        self.drumType = 1
        self.setChannel('10')  # MMA assumes all drums are on channel 10
        self.voice = seqBump([mma.MMA.pat.defaultDrum])
        
    def setVoicing(self, ln):
        """ Set the voicing option for a solo/melody track. Only permitted
            option is FOLLOWCHORD.
        """

        notopt, ln = opt2pair(ln, toupper=True)

        if notopt:
            error("Voicing %s: Each option must be a OPT=VALUE pair." % self.name)

        for opt, val in ln:
            if opt == 'FOLLOWCHORD':
                if val in ('0', 'OFF'):
                    val = None
                elif val in ('1', 'ON'):
                    val = 1
                else:
                    error("%s Voicing FollowChord: expecting On or Off." % self.name)
                self.followChord = val

            elif opt == 'FOLLOWKEY':
                if val in ('0', 'OFF'):
                    val = None
                elif val in ('1', 'ON'):
                    val = 1
                else:
                    error("%s Voicing FollowKey: expecting On or Off." % self.name)

                self.followKey = val

            elif opt == 'ROOT':
                try:
                    self.rootChord = mma.MMA.chords.cdAdjust[val.upper()]
                except KeyError:
                    error("Voicing %s: Chord name %s not valid." % (self.name, val))

            else:
                error("Voicing %s: Only valid options are 'FollowChord', 'FollowKey' and 'Root'."
                      % self.name)

    def formatPattern(self, pat):
        """ Format an existing pattern. Overrides class def! """

        if not pat:
            return ''
        else:
            return pat

    def saveGroove(self, gname):
        """ Save special/local variables for groove. """

        PC.saveGroove(self, gname)  # create storage. Do this 1st.
        self.grooves[gname]['FOLLOWCHORD'] = self.followChord
        self.grooves[gname]['FOLLOWKEY'] = self.followKey
        self.grooves[gname]['ROOT'] = self.rootChord
        self.grooves[gname]['TONES'] = self.toneList[:]

    def restoreGroove(self, gname):
        """ Restore special/local/variables for groove. """

        self.followChord = self.grooves[gname]['FOLLOWCHORD']
        self.followKey = self.grooves[gname]['FOLLOWKEY']
        self.rootChord = self.grooves[gname]['ROOT']
        self.toneList = self.grooves[gname]['TONES']
        PC.restoreGroove(self, gname)

    def setSeqSize(self):
        """ Expand existing pattern list. """

        self.toneList = seqBump(self.toneList)
        PC.setSeqSize(self)

    def clearSequence(self):
        """ Set some initial values. Called from init and clear seq. """

        PC.clearSequence(self)
        self.toneList = seqBump([self.drumTone])

    def defPatRiff(self, ln):
        """ Create a solo pattern. This is used for solo sequenes only.

            All we do is to return the string set as a pattern. Pretty simple.

        """

        return ln

    def setStretch(self, ln):
        """ Set a stretch (or compress) value for each note set in a solo. """

        s = stof(ln[0])

        if s < 1 or s > 500:
            error("%s Stretch: value must be a percentage in range 1 to 500, not '%s'"
                  % (self.name, s))

        self.stretch = s / 100.

    def setArp(self, ln):
        """ Set the arpeggiate options. """

        notopt, ln = opt2pair(ln, 1)

        if notopt:
            error("%s Arpeggiate: expecting cmd=opt pairs, not '%s'."
                  % (self.name, ' '.join(notopt)))

        for cmd, opt in ln:
            if cmd == 'RATE':
                if opt == '0' or opt == 'NONE':
                    self.arpRate = 0
                else:
                    self.arpRate = self.getNoteLen(opt)

            elif cmd == 'DECAY':
                v = stof(opt, "Arpeggiate Decay must be a value, not '%s'" % opt)
                if v < -50 or v > 50:
                    error("%s Arpeggiate: Decay rate must be -50..+50" %
                          self.name)
                self.arpDecay = v / 100

            elif cmd == 'DIRECTION':
                valid = ("UP", "DOWN", "BOTH", "RANDOM")
                if opt not in valid:
                    error("%s Arpeggiate Direction: Unknown setting '%s', use %s."
                          % (self.name, opt, ', '.join(valid)))
                self.arpDirection = opt

        if mma.MMA.debug.debug:
            dPrint("%s Arpeggiate: Rate=%s Decay=%s Direction=%s" % 
                  (self.name, self.arpRate, self.arpDecay, self.arpDirection))

    def getNoteLen(self, n):
        """ This is a special interface to mma.MMA.notelen.getNoteLen() which
            adjusts the length (in MIDI ticks) by the "stetch" value for
            this class.
        """

        return mma.MMA.notelen.getNoteLen(n) * self.stretch

    def restart(self):
        self.ssvoice = -1

    def setTone(self, ln):
        """ A solo track can have a tone, if it is DRUMTYPE."""

        if not self.drumType:
            error("You must set a Solo track to DrumType before setting Tone")

        if len(ln) > 1:
            error("Only 1 value permitted for Drum Tone in Solo tracks")

        self.drumTone = mma.MMA.translate.dtable.get(ln[0])

    def getChord(self, c, velocity, isdrum, isgrace):
        """ Extract a set of notes for a single beat.

            This is a function just to make getLine() a bit shorter
            and more readble.
        """

        c = re.split("[, ]+", c)

        if not c:
            error("You must specify the first note in a solo line")

        """ Convert the note part into a series of midi values
            Notes can be a single note, or a series of notes. And
            each note can be a letter a-g (or r), a '#,&,n' plus
            a series of '+'s or '-'s. Drum solos must have each
            note separated by ' ' or ','s: "Snare1,KickDrum1,44".

            Each chunk could be:
             - a midi value (44)
             - a drum note ( KickDrum1)
             - a single note (g#) (g&-)
             - Or groups with spaces/commas (f 100) (44 , KickDrum) (a,b c)
        """

        events = []   # array for each note event

        for cc in c:
            if not cc or not cc[0]:
                continue
            if '/' in cc:
                if cc.count('/') > 1:
                    error("%s: Only 1 '/velocity' permitted. You can separate "
                          "notes in the chord with ',' or ' ' and it'll work." %
                          self.name)
                cc, newvel = cc.split('/')
                if not newvel:
                    error("%s: expecting 'volume' after '/'" % self.name)

                if not cc:
                    error("%s: Volume '/' must immediately follow note." % self.name)

                thisvel = stoi(newvel)

                if thisvel < 0 or thisvel > 127:
                    error("%s: Velocity must be 0..127, not '%s'." % (self.name, newvel))
            else:
                thisvel = velocity

            if cc[0] == 'r':
                if isgrace:
                    warning("%s: Grace note is a rest (ignored)." % self.name)

                if events or len(cc) > 1:
                    error("%s: Rests and notes cannot be combined." % self.name)
                else:
                    events.append(NoteEvent(None, 0, isgrace))  # note event with no pitch

            elif cc[0] in "1234567890":
                n = stoi(cc, "%s: Note values must be integer or literal." %
                         self.name)
                if n < 0 or n > 127:
                    error("%s: Midi notes must be 0..127, not '%s'" %
                          (self.name, n))

                # if using value we fake-adjust octave,
                # it (and transpose) is set later.

                if not isdrum:
                    n -= self.octave[self.seq]

                events.append(NoteEvent(n, thisvel, isgrace))

            elif isdrum:       # drum must be a value, * or drum-name
                if cc == '*':
                    events.append(NoteEvent(self.drumTone, thisvel, isgrace))
                else:
                    events.append(NoteEvent(int(mma.MMA.translate.dtable.get(cc)), thisvel, isgrace))

            else:   # must be a note(s) in std. notation
                cc = list(cc)
                while cc:
                    name = cc.pop(0)

                    if not name in self.midiNotes:
                        error("%s: Encountered illegal note name '%s'"
                              % (self.name, name))

                    n = self.midiNotes[name]  # name is string, n is value

                    # Parse out a "#', '&' or 'n' accidental.

                    if cc and cc[0] in accValues:
                        i = cc.pop(0)
                        self.acc[name] = accValues[i]

                    n += self.acc[name]  # accidental adjust (from above or keysig)

                    # Parse out +/- (or series) for octave

                    while cc and cc[0] in '+-':
                        a = cc.pop(0)
                        if a == '+':
                            n += 12
                        else:
                            n -= 12

                    events.append(NoteEvent(n, thisvel, isgrace))

        return events

    def getLine(self, pat):
        """ Extract a melodyline for solo/melody tracks.

            This is only called from trackbar(), but it's nicer
            to isolate it here.


            RETURNS: notes structure. This is a dictionary. Each key represents
                     an offset in MIDI ticks in the current bar. The data for
                     each entry is an array of note events:

                 notes[offset] - [nev [,...] ] See top of file for noteEvent()
                                  class which sets the fields.
        """

        sc = self.seq

        """ Get a COPY of the keysignature note table (a dict).
            As a bar is processed the table is updated. There is one difference
            here---in real music an accidental for a note in a given octave does
            not effect the following same-named notes in different octaves.
            In this routine IT DOES.

            FollowKey=Off is usually (user) set when using a sequence pattern.
        """

        if self.followKey:   # this is the default
            self.acc = keySig.accList.copy()
        else:    # disable the feature. Useful for sequence patterns!
            self.acc = {'a': 0, 'c': 0, 'b': 0, 'e': 0, 'd': 0, 'g': 0, 'f': 0}

        # list of notename to midivalues

        self.midiNotes = {'c': 0, 'd': 2, 'e': 4, 'f': 5, 'g': 7, 'a': 9, 'b': 11, 'r': None}

        """ The initial string is in the format "1ab;4c;;4r;". The trailing
            ';' is important and needed. If we don't have this requirement
            we can't tell if the last note is a repeat of the previous. For
            example, if we have coded "2a;2a;" as "2a;;" and we didn't
            have the 'must end with ;' rule, we end up with "2a;" and
            then we make this into 2 notes...or do we? Easiest just to
            insist that all bars end with a ";".
        """

        if not pat.endswith(';'):
            error("All Solo strings must end with a ';'")

        # Get the end of the bar in ticks. If we're doing a truncate
        # use the temporary bar length, otherwise figure it out.

        if mma.MMA.truncate.length:
            barEnd = mma.MMA.truncate.length
        else:
            barEnd = gbl.barLen

        duration = self.getNoteLen('4')    # default note length
        velocity = defaultVelocity         # intial/default velocity for solo notes
        articulation = 1            # additional articulation for solo notes

        notes = {}   # NoteEvent list, keys == offset

        if self.drumType:
            isdrum = 1
            lastc = str(self.drumTone)
        else:
            isdrum = None
            lastc = ''             # last parsed note

        # convert pat to a list
        pat = [x.strip() for x in pat.split(';')[:-1]]

        # set initial offset into bar. This compensates for the previous
        # bar ending in a ~ and this one starting with ~.
        # This special case bumps the initial bar offset

        if pat[0].startswith("~"):
            if not self.endTilde or self.endTilde[1] != gbl.tickOffset:
                error("Previous line did not end with '~'")
            else:
                pat[0] = pat[0][1:].strip()
                offset = self.endTilde[0]
        else:
            offset = 0

        lastOffset = None

        # Strip off trailing ~. This permits long notes to end past the
        # current barend. Note, flag set for the next bar to test for
        # a leading ~.

        if pat[-1].endswith("~"):
            self.endTilde = [1, gbl.tickOffset + barEnd]
            pat[-1] = pat[-1][:-1].strip()
        else:
            self.endTilde = []

        ##################################################
        # Now we can parse each chunk of the solo string.

        for a in pat:
            """ If we find a "<>" we just ignore that. It's useful when
                multiple continuation bars are needed with ~.
            """

            accentVol = None
            accentDur = None
            isgrace = False

            if a == '<>':
                continue

            """ Next, strip out all '<SPECIAL=xx>' settings.

              VOLUME: If no option is set, we assume VOLUME. The default
                      velocity setting was set before the loop (==90) and is
                      changed here for the duration of the current bar/riff.
                      The set velocity will still be modified by the global
                      and track volume adjustments.

              DURATION: Duration or articulation setting is defaulted to 100.
                        Changing it here will do so for the duration of the
                        bar/riff. Note, the track ARTICULATION is still applied.

              ISGRACE: grace note indication and offset. Offset is optional.

              OFFSET: change the current offset into the bar. Can be negative
                      which forces overlapping notes.

            """

            a, vls = pextract(a, "<", ">")

            if vls:
                if len(vls) > 1:
                    error("Only 1 <modifier> is permitted per note-set")

                vls = vls[0].split(',')
                for vv in vls:

                    vv = vv.upper().strip()

                    # We have an default offset for grace notes of 2.
                    # So, if what have a GRACE without a offset, convert
                    # the command to that.
                    if vv == 'GRACE':
                        vv = 'GRACE=2'

                    if not '=' in vv:
                        vv = "VOLUME=" + vv

                    vc, vo = vv.split('=', 1)  # note: it's already uppercase!

                    if vc == 'VOLUME':
                        if vo in mma.MMA.volume.vols:   # arg was a volume 'FF, 'mp', etc.
                            velocity = defaultVelocity * mma.MMA.volume.vols[vo]
                        else:
                            error("%s: No volume '%s'." % (self.name, vo))

                    elif vc == 'GRACE':
                        isgrace = stof(vo, "%s: Expecting a value, not %s." % (self.name, vo))
                        if isgrace <= 0:
                            error("%s: Offset modifier must be greater than 0."
                                  % self.name)

                    elif vc == 'OFFSET':
                        offset = stoi(vo, "%s: Offset expecting integer, not %s."
                                      % (self.name, vo))

                        if offset < 0:
                            error("%s: Offset must be positive." % self.name)

                        if offset >= barEnd:
                            error("%s: Offset has been set past the end of the bar."
                                  % self.name)

                    elif vc == 'ARTICULATE':
                        articulation = stoi(vo, "%s: Articulation expecting integer,"
                                            " not %s." % (self.name, vo))

                        if articulation < 1 or articulation > 200:
                                error("%s: Articulation must be 1..200, not %s." %
                                      (self.name, vo))
                        articulation /= 100.

                    else:
                        error("%s: Unknown command '%s'." % (self.name, vv))

            if offset >= barEnd:
                error("Attempt to start Solo note '%s' after end of bar" % a)

            """ Split the chord chunk into a note length and notes. Each
                part of this is optional and defaults to the previously
                parsed value.
            """

            i = 0
            while i < len(a):
                if not a[i] in '1234567890.+-tT:':
                    break
                else:
                    i += 1

            if i:
                duration = self.getNoteLen(a[0:i].replace(' ', ''))
                a = a[i:].strip()

            # next item might be an accent string.

            i = 0
            while i < len(a):
                if not a[i] in "!-_^&":
                    break
                else:
                    i += 1

            if i:
                c = a[0:i]
                accentVol = 1
                accentDur = 1

                accentDur -= c.count('!') * .2
                accentDur += c.count('-') * .2
                accentDur += c.count('_') * .2
                accentVol += c.count('^') * .2
                accentVol -= c.count('&') * .2

                if accentDur < .1:
                    accentDur = .1
                if accentDur > 2:
                    accentDur = 2
                if accentVol < .1:
                    accentVol = .1
                if accentVol > 2:
                    accentVol = 2
                a = a[i:]

            # Now we get to look at pitches.

            if not a or a == '' or a == ' ':
                a = lastc
            evts = self.getChord(a, velocity, isdrum, isgrace)  # get chord

            if not evts:
                error("No pitches specifed for notes.")

            for e in evts:
                e.velocity = self.adjustVolume(e.defvelocity, offset)
                if accentVol:
                    e.velocity *= accentVol
                e.duration = duration

                if accentDur:
                    e.articulation = articulation * accentDur
                else:
                    e.articulation = articulation

            lastc = a     # save last chord for next loop

            # add note event(s) to note{}
            if not offset in notes:
                notes[offset] = []
            notes[offset].extend(evts)

            lastOffset = offset

            if not e.isgrace:
                offset += duration

        if offset <= barEnd:
            if self.endTilde:
                error("Tilde at end of bar has no effect")

        else:
            if self.endTilde:
                self.endTilde[0] = offset - barEnd
            else:
                warning("%s, end of last note overlaps end of bar by %g "
                        "beat(s)." % (self.name, (offset - barEnd) / float(gbl.BperQ)))

        if mma.MMA.swing.mode:
            notes = mma.MMA.swing.swingSolo(notes)

 
        return notes

    def followAdjust(self, notes, ctable):
        """ Convert pitches to reflect current chord. """

        sc = self.seq

        for offset in notes:
            nn = notes[offset]

            if len(nn) == 1 and nn[0].pitch is not None:
                tb = self.getChordInPos(offset, ctable)

                if tb.chordZ:
                    continue

                nn[0].pitch += (tb.chord.rootNote + self.rootChord)

        return notes

    def addHarmony(self, notes, ctable):
        """ Add harmony to solo notes. We need to be careful
            since the chords can contain grace and NULL notes.
        """

        sc = self.seq

        harmony = self.harmony[sc]
        harmOnly = self.harmonyOnly[sc]

        for offset in notes:
            nn = notes[offset]

            pitch = None
            count = 0
            for n in nn:
                if n.isgrace or n.pitch==None:
                    continue
                pitch = n.pitch
                duration = n.duration
                articulation = n.articulation
                velocity = n.defvelocity
                count += 1  # signals multiple notes, don't harmonize

                if harmOnly:
                    n.pitch = None

            # The chord might have no notes, have more than one, or be all grace
            if pitch is None or count != 1:
                continue

            tb = self.getChordInPos(offset, ctable)

            if tb.chordZ:
                continue

            h = mma.MMA.harmony.harmonize(harmony, pitch, tb.chord.bnoteList)

            for n in h:
                e = NoteEvent(n,
                        self.adjustVolume(velocity * self.harmonyVolume[sc], offset),
                        False)
                e.duration = duration
                e.articulation = articulation
                nn.append(e)

    def trackBar(self, pat, ctable):
        """ Do the solo/melody line. Called from self.bar() """

        sc = self.seq

        notes = self.getLine(pat)

        # adjust all the notes for the given chord if we are in "FOLLOW CHORD" mode.

        if self.followChord and not self.drumType:
            notes = self.followAdjust(notes, ctable)

        if self.harmony[sc] and not self.drumType:
            self.addHarmony(notes, ctable)

        unify = self.unify[sc]

        rptr = self.mallet

        for offset in sorted(notes.keys()):
            nn = notes[offset]
            
            # the "None" test is important. Arp doesn't like rests.
            if self.arpRate and nn[0].pitch is not None:
                self.trackArp(nn, offset)
                continue

            # For each chord we process 2x. First time finds all the grace
            # notes and sends them to the midi machine; 2nd time normal notes.
            # Has to be done this was so we don't force grace note adjustments
            # (durations) onto normal notes. Grace notes ignore strum!

            # grace
            for nev in nn:
                n = nev.pitch
                if n is None or not nev.isgrace:     # skip rests and non-grace
                    continue

                if not self.drumType:        # no octave/transpose for drums
                    n = self.adjustNote(n)

                # Note that each grace note has it's own offset and duration
                self.sendNote(int(offset - (nev.duration // nev.isgrace)),
                              self.getDur(int(nev.duration * nev.articulation)),
                              n, int(nev.velocity))

            # normal notes.

            strumAdjust = self.getStrum(sc)
            strumAdd = self.strumAdd[sc]
            strumOffset = 0

            dur = None  # default duration for notes in chord

            for nev in nn:
                n = nev.pitch
                if n is None or nev.isgrace:     # skip rests and grace
                    continue

                if not self.drumType:        # no octave/transpose for drums
                    n = self.adjustNote(n)

                off = offset + strumOffset
                
                # Set duration for this chord. Only do it once so they are
                # all the same length. The call to getDur() adjusts for RDURATION.
                if dur is None:
                    dur = self.getDur(int(nev.duration * nev.articulation))

                self.sendNote(off, dur, n, int(nev.velocity))

                strumAdjust += strumAdd
                strumOffset += strumAdjust

    def trackArp(self, nn, offset):
        """ Special trackbar() for arpeggiator. """

        if self.drumType:
            error("%s Arpeggiate: Incompatible with DRUMTYPE. Try MALLET?" % self.name)

        notes = [[self.adjustNote(x.pitch), x.velocity] for x in nn]
        notes.sort()

        random = self.direction == 'RANDOM'

        if self.arpDirection == "DOWN":
            notes.reverse()

        elif self.arpDirection == "BOTH":
            z = notes[:]
            z.reverse()
            notes.extend(z[1:-1])

        duration = self.arpRate             # duration of each note
        count = nn[0].duration // duration   # total number to play
        if count < 1:
            count = 1

        while 1:
            nn = range(len(notes))
            if random:
                random.randomize(nn)
            for i in nn:
                n = notes[i]

                self.sendNote(offset,
                              self.getDur(duration), n[0],
                              self.adjustVolume(n[1], offset))
                count -= 1
                if not count:
                    break

                offset += duration

                if self.arpDecay:
                    n[1] = int(n[1] + (n[1] * self.arpDecay))
                    if n[1] < 1:
                        n[1] = 1
                    if n[1] > 127:
                        n[1] = 127

            if not count:
                break


class Solo(Melody):
    """ Pattern class for a solo track. """

    vtype = 'SOLO'

    # Grooves are not saved/restored for solo tracks.

    def restoreGroove(self, gname):
        self.setSeqSize()

    def saveGroove(self, gname):
        pass

    def forceRestoreGroove(self, gname):
        PC.restoreGroove(self, gname)

    def forceSaveGroove(self, gname):
        PC.saveGroove(self, gname)

#######################

""" When solos are included in a chord/data line they are
    assigned to the tracks listed in this list. Users can
    change the tracks with the setAutoSolo command.
"""

autoSoloTracks = ['SOLO', 'SOLO-1', 'SOLO-2', 'SOLO-3']


def setAutoSolo(ln):
    """ Set the order and names of tracks to use when assigning
        automatic solos (specified on chord lines in {}s).
    """

    global autoSoloTracks

    if not len(ln):
        error("You must specify at least one track for autosolos")

    autoSoloTracks = []
    for n in ln:
        n = n.upper()
        mma.MMA.alloc.trackAlloc(n, 1)
        if gbl.tnames[n].vtype not in ('MELODY', 'SOLO'):
            error("All autotracks must be Melody or Solo tracks, not %s"
                  % gbl.tnames[n].vtype)

        autoSoloTracks.append(n)

    if mma.MMA.debug.debug:
        dPrint("AutoSolo track names: %s" % ' '.join([a for a in autoSoloTracks]))


###############

def extractSolo(ln, rptcount):
    """ Parser calls this to extract solo strings. """

    a = ln.count('{')
    b = ln.count('}')

    if a != b:
        error("Mismatched {}s for solo found in chord line")

    if a:
        if rptcount > 1:
            error("Bars with both repeat count and solos are not permitted")

        ln, solo = pextract(ln, '{', '}')

        if len(solo) > len(autoSoloTracks):
            error("Too many melody/solo riffs in chord line. %s used, "
                  "only %s defined" % (len(solo), len(autoSoloTracks)))

        firstSolo = solo[0][:]  # save for autoharmony tracks

        """ We have the solo information. Now we loop though each "solo" and:
              1. Ensure or Create a MMA track for the solo
              2. Push the solo data into a Riff for the given track.
        """

        for s, trk in zip(solo, autoSoloTracks):
            if not s:
                continue    # skip placeholder/empty tracks
            mma.MMA.alloc.trackAlloc(trk, 1)
            t = gbl.tnames[trk]
            if t.riff:
                error("%s: Attempt to add {} solo when the track "
                      "has pending RIFF data." % t.name)
            t.setRiff(s.strip())

        """ After all the solo data is interpreted and sent to the
            correct track, we check any leftover tracks. If any of these
            tracks are  empty of data AND are harmonyonly the note
            data from the first track is interpeted again for that
            track. Tricky: the max() is needed since harmonyonly can
            have different setting for each bar...this way
            the copy is done if ANY bar in the seq has harmonyonly set.
        """

        for t in autoSoloTracks[1:]:
            if t in gbl.tnames and not gbl.tnames[t].riff \
                    and set(gbl.tnames[t].harmonyOnly) != {None}:
                gbl.tnames[t].setRiff(firstSolo[:])

                if mma.MMA.debug.debug:
                    dPrint("%s duplicated to %s for HarmonyOnly." % (trk, t))

    return ln



# readmidi.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

Bob's quick and dirty functions to read a midi file. 
  The entire file is read into memory and then parsed
  into different events.
  
This should work with any midi file. Python 2/3 compatible.
  
Note that many midi events are NOT read. The code here just
  skips over them since they are not useful to mma.MMA.
        
Any and all errors raise a 'RuntimeError'. If you need to, catch
  these from your calling program.
  
  Example:
     Create the class
        mf = MidiData()
     Set some class variables
        mf.velocityAdjust = 200
     Read the file
        try:
          mf.readfile("fever.mid")
        except RuntimeError as e:
          print("ERROR: %s" % e)
        sys.exit(0)
  
The main things to look at after reading the file are:

    events[]      - dict of lists, keys are channel numbers.
                    each item is a list of lists in the format:
                       [offset, event-type, [data] ]
    textEvents[]  - list, each item is [offset, text]
    lyricEvents[] - list, each item is [offset, text]

Other useful info includes: beatDivision, numTracks
  
You might look at beatDivision and adjust the offsets using
  code like this:
     if beatDivision != gbl.BperQ:
        mf.adjustBeats( gbl.BperQ / float(beatDivision))

"""

import sys
from . import gbl

class MidiData:
    def __init__(self):
        # set these before reading the file
        self.ignorePC = True     # skip (don't save) program changes
        self.octaveAdjust = 0    # octave adjustment values (should be -12,24,etc)
        self.velocityAdjust = 100  # Percentage to apply to velocities

        # storage
        self.events = {}
        for c in range(0, 16):
            self.events[c] = []
        self.textEvents = []
        self.lyricEvents = []

        # internal use, not for user
        self.midifile = None
        self.offset = 0

        # useful things to examine after reading
        self.MidiFormat = None
        self.numTracks = None
        self.beatDivision = None
        self.firstNote = None  # offset of 1st note in file

    def mvarlen(self):
        """ Convert variable length midi value to int. """

        x = 0
        for i in range(4):

            try:
                b = self.midifile[self.offset]
                self.offset += 1
            except:
                raise RuntimeError("Invalid MIDI file include (varlen->int)")

            if b < 0x80:
                x = (x << 7) + b
                break
            else:
                x = (x << 7) + (b & 0x7f)

        return int(x)

    def strs(self, count):
        """ Return a string of count chars. """

        s = self.midifile[self.offset:self.offset+count]
        s = s.decode(encoding=gbl.encoding)
        self.offset += count
        return s

    def chars(self, count):
        """ Return 'count' chars from file (updates global pointer). """

        b = list(self.midifile[self.offset:self.offset+count])
        self.offset += count
        return b

    def m1i(self):
        """ Get 1 byte (updates global pointer). """

        try:
            b = self.midifile[self.offset]
            self.offset += 1
        except:
            raise RuntimeError("Invalid MIDI file include (byte, offset=%s)" % self.offset)

        return b


    def m32i(self):
        """ Convert 4 bytes to integer. """

        x = 0
        for i in range(4):
            try:
                b = self.midifile[self.offset]
                self.offset += 1
            except:
                raise ("Invalid MIDI file include (i32->int, offset=%s)" % self.offset)

            x = (x << 8) + b

        return int(x)


    def m16i(self):
        """ Convert 2 bytes to integer. """

        x = 0
        for i in range(2):
            try:
                b = self.midifile[self.offset]
                self.offset += 1
            except:
                raise ("Invalid MIDI file include (i16->int, offset=%s)" % self.offset)
            x = (x << 8) + b

        return int(x)

    def adjustBeats(self, adjustment):
        """ Adjust all the offsets in the generated data to compensate
            between midi timing standards. MMA uses 194 ticks/beat. 
        """

        for ch in self.events:
            for e in self.events[ch]:
                e[0] = int(e[0] * adjustment)
        for e in self.textEvents:
            e[0] = int(e[0] * adjustment)
        for e in self.lyricEvents:
            e[0] = int(e[0] * adjustment)

        self.beatDivision = int(self.beatDivision * adjustment)
        self.firstNote = int(self.firstNote * adjustment)

        return None

    ### Main reader

    def readFile(self, filename):
        """ Read the midi file into memory and parse it. """

        try:
            inpath = open(filename, "rb")
        except:
            raise RuntimeError("Unable to open MIDI file '%s' for reading." % filename )

        try:
            self.midifile = inpath.read()
        except:
            raise RuntimeError("Unable to read MIDI file '%s'." % filename)

        self.midifile = bytearray(self.midifile)  # needed for python2, does nothing in 3
        inpath.close()

        # legit midi file?
        
        hd = self.strs(4)
        if hd != 'MThd':
            raise RuntimeError("Expecting 'MThd', %s not a standard midi file" % filename)

        if self.m32i() != 6:
            raise RuntimeError("Expecting a 32 bit value of 6 in header")

        self.MidiFormat = self.m16i()

        if self.MidiFormat not in (0, 1):
            raise ("MIDI file format %s not recognized" % self.MidiFormat)

        self.numTracks = self.m16i()
        self.beatDivision = self.m16i()

        self.firstNote = 0xffffff

        # Now read the data, track by track

        for tr in range(self.numTracks):
            tm = 0

            # Validity for each track

            hdr = self.strs(4)
            if hdr != 'MTrk':
                raise RuntimeError("Malformed MIDI file in track header")
            trlen = self.m32i()  # track length, not used ... advances ptr

            lastevent = None

            while 1:
                tm += self.mvarlen()        # adjust total offset by delta

                ev = self.m1i()

                if ev < 0x80:
                    if not lastevent:
                        raise RuntimeError ("Illegal running status in %s at %s" %
                              (self.midifile, self.offset))
                    self.offset -= 1
                    ev = lastevent

                sValue = ev >> 4        # Shift MSBs to get a 4 bit value 
                channel = ev & 0x0f

                if sValue == 0x8:        # note off event
                    note = self.m1i()
                    vel = self.m1i()

                    if self.octaveAdjust and channel != 9:  # drums are 9 when 0..15 (not 10!)
                        note += self.octaveAdjust
                        while note < 0:
                            note += 12
                        while note > 127:
                            note -= 12
                    self.events[channel].append([tm, ev & 0xf0, note, vel])

                elif sValue == 0x9:        # note on event
                    if tm < self.firstNote:
                        self.firstNote = tm
                    note = self.m1i()
                    vel = self.m1i()

                    if self.octaveAdjust and channel != 9:
                        note += self.octaveAdjust
                        while note < 0:
                            note += 12
                        while note > 127:
                            note -= 12

                    if self.velocityAdjust != 100:
                        vel = (vel * self.velocityAdjust) // 100
                        if vel < 0:
                            vel = 1
                        if vel > 127:
                            vel = 127

                    self.events[ev & 0xf].append([tm, ev & 0xf0, note, vel])

                elif sValue == 0xa:        # key pressure
                    self.events[ev & 0xf].append([tm, ev & 0xf0, self.chars(2)])

                elif sValue == 0xb:        # control change
                    self.events[ev & 0xf].append([tm, ev & 0xf0, self.chars(2)])

                elif sValue == 0xc:        # program change
                    if self.ignorePC:      # default is to ignore these
                        self.offset += 1
                    else:                  # set with option ignorePC=1
                        self.events[ev & 0xf].append([tm, ev & 0xf0, self.chars(1)])

                elif sValue == 0xd:        # channel pressure
                    self.events[ev & 0xf].append([tm, ev & 0xf0, self.chars(1)])

                elif sValue == 0xe:        # pitch blend
                    self.events[ev & 0xf].append([tm, ev & 0xf0, self.chars(2)])

                elif sValue == 0xf:       # system, mostly ignored
                    if ev == 0xff:        # meta events
                        a = self.m1i()

                        if a == 0x00:    # sequence number
                            l = self.mvarlen()
                            self.offset += l

                        elif a == 0x01:  # text (could be lyrics)
                            self.textEvents.append([tm, self.strs(self.mvarlen())])
  
                        elif a == 0x02:  # copyright
                            l = self.mvarlen()
                            self.offset += l

                        elif a == 0x03:  # seq/track name
                            l = self.mvarlen()
                            self.offset += l

                        elif a == 0x04:  # instrument name
                            l = self.mvarlen()
                            self.offset += l

                        elif a == 0x05:  # lyric
                            self.lyricEvents.append([tm, self.strs(self.mvarlen())])

                        elif a == 0x06:  # marker
                            l = self.mvarlen()
                            self.offset += l

                        elif a == 0x07:  # cue point
                            l = self.mvarlen()
                            self.offset += l

                        elif a == 0x21:  # midi port
                            l = self.mvarlen()
                            self.offset += l

                        elif a == 0x2f:  # end of track
                            l = self.mvarlen()
                            self.offset += l
                            break

                        elif a == 0x51:  # tempo
                            l = self.mvarlen()
                            self.offset += l

                        elif a == 0x54:  # SMPTE offset
                            l = self.mvarlen()
                            self.offset += l

                        elif a == 0x58:  # time sig
                            l = self.mvarlen()
                            self.offset += l

                        elif a == 0x59:  # key sig
                            l = self.mvarlen()
                            self.offset += l

                        else:        # probably 0x7f, proprietary event
                            l = self.mvarlen()
                            self.offset += l

                    elif ev == 0xf0:    # system exclusive
                        l = self.mvarlen()
                        self.offset += l

                    elif ev == 0xf2:    # song position pointer, 2 bytes
                        self.offset += 2

                    elif ev == 0xf3:    # song select, 1 byte
                        self.offset += 1

                    else:        # all others are single byte commands
                        pass

                if ev >= 0x80 and ev <= 0xef:
                    lastevent = ev

        return None



# main.py

"""
The program "MMA - Musical Midi Accompaniment" and the associated
modules distributed with it are protected by copyright.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

import os

import mma.MMA.midi
import mma.MMA.midifuncs
import mma.MMA.parse
import mma.MMA.file
import mma.MMA.options
import mma.MMA.auto
import mma.MMA.docs
import mma.MMA.tempo
import mma.MMA.debug
from  mma.MMA.safe_eval import safeEnv

from . import gbl
from   mma.MMA.common import *
from   mma.MMA.lyric import lyric

import mma.MMA.paths

cmdSMF = None

########################################
########################################

# This is the program mainline. It is called/executed
# exactly once from a call in the stub program mma.MMA.py.

# for some reason, someone might want a different encoding
# real easy to set it from env at startup
m = safeEnv( 'MMA_ENCODING' )
if m:    # don't set to empty ... will crash
    gbl.encoding = m

# MMA prints errors/warning/debug to stdout
# this will redirect to a file
gbl.logFile = safeEnv('MMA_LOGFILE')

mma.MMA.paths.init()   # initialize the lib/include paths


# Get our command line stuff

mma.MMA.options.opts()

#  LibPath and IncPath are set before option parsing, but
#  debug setting wasn't. So we need to do the debug for this now
if mma.MMA.debug.debug:
    dPrint("Initialization has set LibPath set to %s" % mma.MMA.paths.libPath)
    dPrint("Initialization has set IncPath set to %s" % mma.MMA.paths.incPath)

#######################################
# Set up initial meta track stuff. Track 0 == meta

m = gbl.mtrks[0] = mma.MMA.midi.Mtrk(0)

if gbl.infile:
    if gbl.infile != 1:
        fileName = mma.MMA.file.locFile(gbl.infile, None)
        if fileName and not gbl.noCredit:
            m.addTrkName(0, "%s" % fileName.rstrip(".mma"))
            m.addText(0, "Created by mma.MMA. Input filename: %s" % fileName)


m.addTempo(0, gbl.tempo)      # most user files will override this
mma.MMA.tempo.setTime(['4/4'])    # and this. IMPORTANT! Sets default chordTabs[]
   
# Read RC files
mma.MMA.paths.readRC()


################################################
# Update the library database file(s) (-g option)
# Note: This needs to be here, after reading of RC files

if gbl.makeGrvDefs:
    if gbl.infile:
        error("No filename is permitted with the -g option")
    mma.MMA.auto.libUpdate()                # update and EXIT


################################
# We need an input file for anything after this point.

if not gbl.infile:
    mma.MMA.options.usage("No input filename specified.")

################################
# Just extract docs (-Dxh, etc) to stdout.

if gbl.createDocs:
    if gbl.createDocs == 4:
        mma.MMA.docs.htmlGraph(gbl.infile)
    else:
        f = mma.MMA.file.locFile(gbl.infile, None)
        if not f:
            error("File '%s' not found" % gbl.infile)
        mma.MMA.parse.parseFile(f)
        mma.MMA.docs.docDump()
    sys.exit(0)


#########################################################
# This cmdline option overrides the setting in RC files

if mma.MMA.options.cmdSMF is not None:
    gbl.lineno = -1
    mma.MMA.midifuncs.setMidiFileType(['SMF=%s' % mma.MMA.options.cmdSMF])

######################################
# Create the output filename

if not mma.MMA.debug.noOutput:
    mma.MMA.paths.createOutfileName(".mid")


################################################
# Read/process files....

# First the mmastart files
mma.MMA.paths.dommaStart()

# The song file specified on the command line

if gbl.infile == 1:  # use stdin, set filename to 1
    f = 1
else:
    f = mma.MMA.file.locFile(gbl.infile, None)

    if not f:
        gbl.lineno = -1
        error("Input file '%s' not found" % gbl.infile)

mma.MMA.parse.parseFile(f)

# Finally, the mmaend files
mma.MMA.paths.dommaEnd() 

#################################################
# Just display the channel assignments (-c) and exit...

if mma.MMA.debug.chshow:
    msg = ["\nFile '%s' parsed, but no MIDI file produced!" % gbl.infile]
    msg.append("\nTracks allocated:\n")
    k = list(gbl.tnames.keys())
    k.sort()
    max = 0
    for a in k + gbl.deletedTracks:
        if len(a) > max:
            max = len(a)
    max += 1
    wrap = 0
    for a in k:
        wrap += max
        if wrap > 60:
            wrap = max
            msg.append('\n')
        msg.append(" %-*s" % (max, a))
    msg.append('\n')
    print(' '.join(msg))
    
    if gbl.deletedTracks:
        msg = ["Deleted Tracks:\n"]
        wrap = 0
        for a in gbl.deletedTracks:
            wrap += max
            if wrap > 60:
                wrap = max
                msg.append('\n')
            msg.append(" %-*s" % (max, a),)
        msg.append('\n')
        print(' '.join(msg))
        
    msg=["Channel assignments:\n"]
    for c, n in sorted(gbl.midiAssigns.items()):
        if n:
            wrap = 3
            msg.append(" %2s" % c)
            for nn in n:
                wrap += max
                if wrap > 63:
                    msg.append('\n    ')
                    wrap = max+3
                msg.append(" %-*s" % (max, nn))
            msg.append('\n')
    print(' '.join(msg))

    sys.exit(0)


####################################
# Dry run, no output

if mma.MMA.debug.noOutput:
    gbl.lineno = -1
    warning("Input file parsed successfully. No midi file generated")
    sys.exit(0)


##############################
# Create the output (MIDI) file

gbl.lineno = -1    # disable line nums for error/warning


""" We fix the outPath now. This lets you set outpath in the song file.

    The filename "outfile" was created in paths, get a copy.

    It is either the input filename with '.mma' changed to '.mid' (or kar)
    OR if -f<FILE> was used then it's just <FILE>.

    If any of the following is true we skip inserting the outputpath into the
    filename:

        - if outfile starts with a '/'
        - if outPath was not set
        - if -f was used

    Next, the outPath is inserted into the filename. If outPath starts with
    a ".", "/" or "\ " then it is inserted at the start of the path;
    otherwise it is inserted before the filename portion.
"""

outfile = mma.MMA.paths.outfile

if (not outfile.startswith('/')) and gbl.outPath \
        and not gbl.outfile and not gbl.playFile:
    if gbl.outPath[0] in '.\\/':
        outfile = "%s/%s" % (gbl.outPath, outfile)
    else:
        head, tail = os.path.split(outfile)
        outfile = "%s/%s/%s" % (head, gbl.outPath, tail)

fileExist = os.path.exists(outfile)

""" Check if any pending midi events are still around. Mostly
    this will be a DRUM event which was assigned to the 'DRUM'
    track, but no DRUM track was used, just DRUM-xx tracks used.
"""

for n in gbl.tnames.values():
    if n.channel:
        n.clearPending()
        n.doMidiClear()
        n.doChannelReset()
        if n.riff:
            warning("%s has pending Riff(s)" % n.name)

""" Check all the tracks and find total number used. When
    initializing each track (class) we made an initial entry
    in the track at offset 0 for the track name, etc. So, if the
    track only has one entry we can safely skip the entire track.
"""

trackCount = 1    # account for meta track

for n in sorted(gbl.mtrks.keys())[1:]:     # check all but 0 (meta)
    if len(gbl.mtrks[n].miditrk) > 1:
        trackCount += 1

if gbl.printProcessed:
    import mma.MMA.rangeify
    print ("Bars processed: %s" % mma.MMA.rangeify.rangeify(gbl.barLabels))

if trackCount == 1:  # only meta track
    if fileExist:
        print("\n")
    print("No data created. Did you remember to set a groove/sequence?")
    if fileExist:
        print("Existing file '%s' has not been modified." % outfile)
    sys.exit(1)

lyric.leftovers()

if fileExist:
    msg = "Overwriting existing"
else:
    msg = "Creating new"

print("%s midi file (%s bars, %.2f min / %d:%02d m:s): '%s'" %
    (msg, gbl.barNum, gbl.totTime, gbl.totTime, (gbl.totTime%1)*60, outfile))

# Insert the estimated play time in seconds into a comment line.
# A player program can search for this and display it. The value
# will be terminated by a null which should be easy enuf to search for.

gbl.mtrks[0].addText(0, "DURATION: %d" % (gbl.totTime*60) )

try:
    out = open(outfile, 'wb')
except IOError:
    error("Can't open file '%s' for writing" % outfile)

mma.MMA.midi.writeTracks(out)
out.close()

if gbl.playFile:
    import mma.MMA.player
    mma.MMA.player.playMidi(outfile)

if mma.MMA.debug.debug:
    dPrint("Completed processing file '%s'." % outfile)

